# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import uuidutils
import six

from sahara import conductor as c
from sahara import context
from sahara import exceptions
from sahara.i18n import _
from sahara.plugins import base as plugin_base
from sahara.plugins import utils as u
from sahara.service.edp import job_manager
from sahara.service.edp.utils import shares
from sahara.service.health import verification_base as ver_base
from sahara.service import ntp_service
from sahara.service import trusts
from sahara.utils import cluster as c_u
from sahara.utils.openstack import nova
from sahara.utils import remote
from sahara.utils import rpc as rpc_utils

conductor = c.API
CONF = cfg.CONF
LOG = logging.getLogger(__name__)


INFRA = None


def setup_ops(engine):
    global INFRA

    INFRA = engine


class LocalOps(object):
    def provision_cluster(self, cluster_id):
        context.spawn("cluster-creating-%s" % cluster_id,
                      _provision_cluster, cluster_id)

    def provision_scaled_cluster(self, cluster_id, node_group_id_map,
                                 node_group_instance_map=None):
        context.spawn("cluster-scaling-%s" % cluster_id,
                      _provision_scaled_cluster, cluster_id, node_group_id_map,
                      node_group_instance_map)

    def terminate_cluster(self, cluster_id, force=False):
        context.spawn("cluster-terminating-%s" % cluster_id,
                      terminate_cluster, cluster_id, force)

    def run_edp_job(self, job_execution_id):
        context.spawn("Starting Job Execution %s" % job_execution_id,
                      _run_edp_job, job_execution_id)

    def cancel_job_execution(self, job_execution_id):
        context.spawn("Canceling Job Execution %s" % job_execution_id,
                      _cancel_job_execution, job_execution_id)

    def delete_job_execution(self, job_execution_id):
        context.spawn("Deleting Job Execution %s" % job_execution_id,
                      _delete_job_execution, job_execution_id)

    def handle_verification(self, cluster_id, values):
        context.spawn('Handling Verification for cluster %s' % cluster_id,
                      _handle_verification, cluster_id, values)

    def get_engine_type_and_version(self):
        return INFRA.get_type_and_version()

    def job_execution_suspend(self, job_execution_id):
        context.spawn("Suspend Job Execution %s" % job_execution_id,
                      _suspend_job_execution, job_execution_id)


class RemoteOps(rpc_utils.RPCClient):
    def __init__(self):
        target = messaging.Target(topic='sahara-ops', version='1.0')
        super(RemoteOps, self).__init__(target)

    def provision_cluster(self, cluster_id):
        self.cast('provision_cluster', cluster_id=cluster_id)

    def update_keypair(self, cluster_id):
        self.cast('update_keypair', cluster_id=cluster_id)

    def provision_scaled_cluster(self, cluster_id, node_group_id_map,
                                 node_group_instance_map=None):
        self.cast('provision_scaled_cluster', cluster_id=cluster_id,
                  node_group_id_map=node_group_id_map,
                  node_group_instance_map=node_group_instance_map)

    def terminate_cluster(self, cluster_id, force=False):
        self.cast('terminate_cluster', cluster_id=cluster_id, force=force)

    def run_edp_job(self, job_execution_id):
        self.cast('run_edp_job', job_execution_id=job_execution_id)

    def cancel_job_execution(self, job_execution_id):
        self.cast('cancel_job_execution',
                  job_execution_id=job_execution_id)

    def delete_job_execution(self, job_execution_id):
        self.cast('delete_job_execution',
                  job_execution_id=job_execution_id)

    def handle_verification(self, cluster_id, values):
        self.cast('handle_verification', cluster_id=cluster_id, values=values)

    def get_engine_type_and_version(self):
        return self.call('get_engine_type_and_version')

    def job_execution_suspend(self, job_execution_id):
        self.cast('job_execution_suspend', job_execution_id=job_execution_id)


def request_context(func):
    @functools.wraps(func)
    def wrapped(self, ctx, *args, **kwargs):
        context.set_ctx(context.Context(**ctx))
        return func(self, *args, **kwargs)

    return wrapped


class OpsServer(rpc_utils.RPCServer):
    def __init__(self):
        target = messaging.Target(topic='sahara-ops',
                                  server=uuidutils.generate_uuid(),
                                  version='1.0')
        super(OpsServer, self).__init__(target)

    @request_context
    def provision_cluster(self, cluster_id):
        _provision_cluster(cluster_id)

    @request_context
    def update_keypair(self, cluster_id):
        _update_keypair(cluster_id)

    @request_context
    def provision_scaled_cluster(self, cluster_id, node_group_id_map,
                                 node_group_instance_map=None):
        _provision_scaled_cluster(cluster_id, node_group_id_map,
                                  node_group_instance_map)

    @request_context
    def terminate_cluster(self, cluster_id, force=False):
        terminate_cluster(cluster_id, force)

    @request_context
    def run_edp_job(self, job_execution_id):
        _run_edp_job(job_execution_id)

    @request_context
    def cancel_job_execution(self, job_execution_id):
        _cancel_job_execution(job_execution_id)

    @request_context
    def delete_job_execution(self, job_execution_id):
        _delete_job_execution(job_execution_id)

    @request_context
    def handle_verification(self, cluster_id, values):
        _handle_verification(cluster_id, values)

    @request_context
    def get_engine_type_and_version(self):
        return INFRA.get_type_and_version()

    @request_context
    def job_execution_suspend(self, job_execution_id):
        _suspend_job_execution(job_execution_id)


def _setup_trust_for_cluster(cluster):
    cluster = conductor.cluster_get(context.ctx(), cluster)
    trusts.create_trust_for_cluster(cluster)
    trusts.use_os_admin_auth_token(cluster)


def ops_error_handler(description):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(cluster_id, *args, **kwds):
            ctx = context.ctx()
            try:
                # Clearing status description before executing
                c_u.change_cluster_status_description(cluster_id, "")
                f(cluster_id, *args, **kwds)
            except Exception as ex:
                # something happened during cluster operation
                cluster = conductor.cluster_get(ctx, cluster_id)
                # check if cluster still exists (it might have been removed)
                if (cluster is None or
                        cluster.status == c_u.CLUSTER_STATUS_DELETING):
                    LOG.debug("Cluster was deleted or marked for deletion. "
                              "Canceling current operation.")
                    return

                msg = six.text_type(ex)
                LOG.exception("Error during operating on cluster (reason: "
                              "{reason})".format(reason=msg))

                try:
                    # trying to rollback
                    desc = description.format(reason=msg)
                    if _rollback_cluster(cluster, ex):
                        c_u.change_cluster_status(
                            cluster, c_u.CLUSTER_STATUS_ACTIVE, desc)
                    else:
                        c_u.change_cluster_status(
                            cluster, c_u.CLUSTER_STATUS_ERROR, desc)
                except Exception as rex:
                    cluster = conductor.cluster_get(ctx, cluster_id)
                    # check if cluster still exists (it might have been
                    # removed during rollback)
                    if (cluster is None or
                            cluster.status == c_u.CLUSTER_STATUS_DELETING):
                        LOG.debug("Cluster was deleted or marked for deletion."
                                  " Canceling current operation.")
                        return

                    LOG.exception(
                        "Error during rollback of cluster (reason:"
                        " {reason})".format(reason=six.text_type(rex)))
                    desc = "{0}, {1}".format(msg, six.text_type(rex))
                    c_u.change_cluster_status(
                        cluster, c_u.CLUSTER_STATUS_ERROR,
                        description.format(reason=desc))
        return wrapper
    return decorator


def _rollback_cluster(cluster, reason):
    _setup_trust_for_cluster(cluster)
    context.set_step_type(_("Engine: rollback cluster"))
    return INFRA.rollback_cluster(cluster, reason)


def _prepare_provisioning(cluster_id):
    ctx = context.ctx()
    cluster = conductor.cluster_get(ctx, cluster_id)
    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)

    for nodegroup in cluster.node_groups:
        update_dict = {}
        update_dict["image_username"] = INFRA.get_node_group_image_username(
            nodegroup)
        conductor.node_group_update(ctx, nodegroup, update_dict)

    _setup_trust_for_cluster(cluster)

    cluster = conductor.cluster_get(ctx, cluster_id)

    return ctx, cluster, plugin


def _update_sahara_info(ctx, cluster):
    sahara_info = {
        'infrastructure_engine': INFRA.get_type_and_version(),
        'remote': remote.get_remote_type_and_version()}

    return conductor.cluster_update(
        ctx, cluster, {'sahara_info': sahara_info})


@ops_error_handler(
    _("Creating cluster failed for the following reason(s): {reason}"))
def _provision_cluster(cluster_id):
    ctx, cluster, plugin = _prepare_provisioning(cluster_id)

    cluster = _update_sahara_info(ctx, cluster)

    # updating cluster infra
    cluster = c_u.change_cluster_status(
        cluster, c_u.CLUSTER_STATUS_INFRAUPDATING)
    plugin.update_infra(cluster)

    # creating instances and configuring them
    cluster = conductor.cluster_get(ctx, cluster_id)
    context.set_step_type(_("Engine: create cluster"))
    INFRA.create_cluster(cluster)
    ntp_service.configure_ntp(cluster_id)

    # configure cluster
    cluster = c_u.change_cluster_status(
        cluster, c_u.CLUSTER_STATUS_CONFIGURING)
    context.set_step_type(_("Plugin: configure cluster"))
    if hasattr(plugin, 'validate_images'):
        plugin.validate_images(cluster, test_only=False)
    shares.mount_shares(cluster)
    plugin.configure_cluster(cluster)

    # starting prepared and configured cluster
    cluster = c_u.change_cluster_status(
        cluster, c_u.CLUSTER_STATUS_STARTING)

    context.set_step_type(_("Plugin: start cluster"))
    plugin.start_cluster(cluster)

    # cluster is now up and ready
    cluster = c_u.change_cluster_status(
        cluster, c_u.CLUSTER_STATUS_ACTIVE)

    # schedule execution pending job for cluster
    for je in conductor.job_execution_get_all(ctx, cluster_id=cluster.id):
        job_manager.run_job(je.id)

    _refresh_health_for_cluster(cluster_id)


def _specific_inst_to_delete(node_group, node_group_instance_map=None):
    if node_group_instance_map:
        if node_group.id in node_group_instance_map:
            return True
    return False


@ops_error_handler(
    _("Scaling cluster failed for the following reason(s): {reason}"))
def _provision_scaled_cluster(cluster_id, node_group_id_map,
                              node_group_instance_map=None):
    """Provision scaled cluster.

    :param cluster_id: Id of cluster to be scaled.

    :param node_group_id_map: Dictionary in the format
                   node_group_id: number of instances.

    :param node_group_instance_map: Specifies the instances to be removed in
                   each node group.
    """
    ctx, cluster, plugin = _prepare_provisioning(cluster_id)

    # Decommissioning surplus nodes with the plugin
    cluster = c_u.change_cluster_status(
        cluster, c_u.CLUSTER_STATUS_DECOMMISSIONING)

    try:
        instances_to_delete = []
        for node_group in cluster.node_groups:
            ng_inst_to_delete_count = 0
            # new_count is the new number of instance on the current node group
            new_count = node_group_id_map[node_group.id]
            if new_count < node_group.count:
                # Adding selected instances to delete to the list
                if _specific_inst_to_delete(node_group,
                                            node_group_instance_map):
                    for instance_ref in node_group_instance_map[node_group.id]:
                        instances_to_delete.append(_get_instance_obj(
                            node_group.instances, instance_ref))
                        ng_inst_to_delete_count += 1

                # Adding random instances to the list when the number of
                # specific instances does not equals the difference between the
                # current count and the new count of instances.
                while node_group.count - new_count > ng_inst_to_delete_count:
                    instances_to_delete.append(_get_random_instance_from_ng(
                        node_group.instances, instances_to_delete))
                    ng_inst_to_delete_count += 1

        if instances_to_delete:
            context.set_step_type(_("Plugin: decommission cluster"))
            plugin.decommission_nodes(cluster, instances_to_delete)

        # Scaling infrastructure
        cluster = c_u.change_cluster_status(
            cluster, c_u.CLUSTER_STATUS_SCALING)
        context.set_step_type(_("Engine: scale cluster"))
        instance_ids = INFRA.scale_cluster(cluster, node_group_id_map,
                                           instances_to_delete)
        # Setting up new nodes with the plugin
        if instance_ids:
            ntp_service.configure_ntp(cluster_id, instance_ids)
            cluster = c_u.change_cluster_status(
                cluster, c_u.CLUSTER_STATUS_CONFIGURING)
            instances = c_u.get_instances(cluster, instance_ids)
            context.set_step_type(_("Plugin: scale cluster"))
            plugin.scale_cluster(cluster, instances)

        c_u.change_cluster_status(cluster, c_u.CLUSTER_STATUS_ACTIVE)
        _refresh_health_for_cluster(cluster_id)

    except Exception as e:
        c_u.change_cluster_status(cluster, c_u.CLUSTER_STATUS_ACTIVE,
                                  six.text_type(e))


def _get_instance_obj(instances, instance_ref):
    for instance in instances:
        if (instance.instance_id == instance_ref or
                instance.instance_name == instance_ref):
            return instance

    raise exceptions.NotFoundException(str(instance_ref),
                                       _("Instance %s not found"))


def _get_random_instance_from_ng(instances, instances_to_delete):
    # instances list doesn't order by creating date, so we should
    # sort it to make sure deleted instances same as heat deleted.
    insts = sorted(instances,
                   key=lambda x: int(x['instance_name'].split('-')[-1]))
    for instance in reversed(insts):
        if instance not in instances_to_delete:
            return instance


@ops_error_handler(
    _("Terminating cluster failed for the following reason(s): {reason}"))
def terminate_cluster(cluster_id, force=False):
    ctx = context.ctx()
    _setup_trust_for_cluster(cluster_id)

    job_manager.update_job_statuses(cluster_id=cluster_id)
    cluster = conductor.cluster_get(ctx, cluster_id)

    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)
    context.set_step_type(_("Plugin: shutdown cluster"))
    plugin.on_terminate_cluster(cluster)

    context.set_step_type(_("Engine: shutdown cluster"))
    INFRA.shutdown_cluster(cluster, force)

    trusts.delete_trust_from_cluster(cluster)

    conductor.cluster_destroy(ctx, cluster)


def _run_edp_job(job_execution_id):
    job_manager.run_job(job_execution_id)


def _suspend_job_execution(job_execution_id):
    job_manager.suspend_job(job_execution_id)


def _cancel_job_execution(job_execution_id):
    job_manager.cancel_job(job_execution_id)


def _delete_job_execution(job_execution_id):
    try:
        job_execution = job_manager.cancel_job(job_execution_id)
        if not job_execution:
            # job_execution was deleted already, nothing to do
            return
    except exceptions.CancelingFailed:
        LOG.error("Job execution can't be cancelled in time. "
                  "Deleting it anyway.")
    conductor.job_execution_destroy(context.ctx(), job_execution_id)


def _refresh_health_for_cluster(cluster_id):
    st_dict = {'verification': {'status': 'START'}}
    try:
        ver_base.validate_verification_start(cluster_id)
        ver_base.handle_verification(cluster_id, st_dict)
    except ver_base.CannotVerifyError:
        LOG.debug("Cannot verify cluster because verifications are disabled "
                  "or cluster already is verifying")
    except Exception:
        # if occasional error occurred, there is no reason to move
        # cluster into error state
        LOG.debug("Skipping refreshing cluster health")
        ver_base.clean_verification_data(cluster_id)


def _handle_verification(cluster_id, values):
    ver_base.handle_verification(cluster_id, values)


def _update_keypair(cluster_id):
    ctx = context.ctx()
    cluster = conductor.cluster_get(ctx, cluster_id)
    keypair_name = cluster.user_keypair_id
    key = nova.get_keypair(keypair_name)
    nodes = u.get_instances(cluster)
    for node in nodes:
        with node.remote() as r:
            r.execute_command(
                "echo {keypair} >> ~/.ssh/authorized_keys".
                format(keypair=key.public_key))



# Copyright (c) 2017 Massachusetts Open Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import six.moves.urllib.parse as urlparse

import sahara.exceptions as ex
from sahara.i18n import _
from sahara.service.edp.job_binaries.base import JobBinaryType
from sahara.service.edp import s3_common


class S3Type(JobBinaryType):
    def copy_binary_to_cluster(self, job_binary, **kwargs):
        r = kwargs.pop('remote')

        dst = self._generate_valid_path(job_binary)
        raw = self.get_raw_data(job_binary)

        r.write_file_to(dst, raw)
        return dst

    def validate_job_location_format(self, url):
        url = urlparse.urlparse(url)
        return url.scheme == "s3" and url.hostname

    def validate(self, data, **kwargs):
        # We only check on create, not update
        if not kwargs.get('job_binary_id', None):
            s3_common._validate_job_binary_url(data['url'])
            extra = data.get("extra", {})
            if (six.viewkeys(extra) !=
                    {"accesskey", "secretkey", "endpoint"}):
                raise ex.InvalidDataException(
                    _("Configs 'accesskey', 'secretkey', and 'endpoint'"
                      " must be provided."))

    def get_raw_data(self, job_binary, **kwargs):
        return s3_common.get_raw_job_binary_data(job_binary)



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import uuidutils
import six.moves.urllib.parse as urlparse

from sahara import conductor as c
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.service.edp.job_binaries.base import JobBinaryType
from sahara.service.edp import job_utils
from sahara.service.edp.utils import shares as shares_service
from sahara.utils.openstack import manila as m

conductor = c.API


class ManilaType(JobBinaryType):
    def copy_binary_to_cluster(self, job_binary, **kwargs):
        remote = kwargs.pop('remote')
        path = self._get_share_path(job_binary, remote)
        # if path is None, then it was mounted in the default path
        # by the prepare_cluster method so just construct the
        # default path and return it
        if path is None:
            url = urlparse.urlparse(job_binary.url)
            share_id = url.netloc
            mount_point = shares_service.default_mount(share_id)
            path = shares_service.make_share_path(mount_point, url.path)
        return path

    def prepare_cluster(self, job_binary, **kwargs):
        remote = kwargs.pop('remote')
        path = self._get_share_path(job_binary, remote)
        if path is None:
            path = job_utils.mount_share_at_default_path(
                job_binary.url, remote.instance.node_group.cluster)

    def _get_share_path(self, job_binary, remote):
        shares = []
        if remote.instance.node_group.cluster.shares:
            shares.extend(remote.instance.node_group.cluster.shares)
        if remote.instance.node_group.shares:
            shares.extend(remote.instance.node_group.shares)

        path = shares_service.get_share_path(job_binary.url, shares)
        return path

    def validate_job_location_format(self, url):
        if url.startswith(m.MANILA_PREFIX):
            url = urlparse.urlparse(url)
            return (uuidutils.is_uuid_like(url.netloc) and url.path)
        else:
            return False

    def validate(self, data, **kwargs):
        self._validate_url(data['url'])

    def _validate_url(self, url):
        if len(url) == 0:
            raise ex.InvalidDataException(_("Manila url must not be empty"))
        url = urlparse.urlparse(url)
        if url.scheme != "manila":
            raise ex.InvalidDataException(_("Manila url scheme must be"
                                            " 'manila'"))
        if not uuidutils.is_uuid_like(url.netloc):
            raise ex.InvalidDataException(_("Manila url netloc must be a"
                                            " uuid"))
        if not url.path:
            raise ex.InvalidDataException(_("Manila url path must not be"
                                            " empty"))

    def get_raw_data(self, job_binary, **kwargs):
        raise ex.NotImplementedException('Manila does not implement this '
                                         'method')



# Copyright (c) 2016 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from sahara import context
from sahara.db.sqlalchemy import api
from sahara.db.sqlalchemy import models as m
import sahara.tests.unit.base as base


class TestPaginationUtils(testtools.TestCase):

    def test_get_prev_and_next_objects(self):

        query = [mock.MagicMock(id=i) for i in range(100)]

        res = api._get_prev_and_next_objects(query, 5, None)
        self.assertEqual((None, 4), res)

        res = api._get_prev_and_next_objects(query, None, None)
        self.assertEqual((None, None), res)

        res = api._get_prev_and_next_objects(query, 5, mock.MagicMock(id=42))
        self.assertEqual((37, 47), res)

        res = api._get_prev_and_next_objects(query, 5, mock.MagicMock(id=4))
        self.assertEqual((None, 9), res)

        res = api._get_prev_and_next_objects(query, 5, mock.MagicMock(id=100))
        self.assertEqual((None, None), res)

    def test_parse_sorting_args(self):
        self.assertEqual(("name", "desc"), api._parse_sorting_args("-name"))
        self.assertEqual(("name", "asc"), api._parse_sorting_args("name"))


class TestRegex(testtools.TestCase):

    def test_get_regex_op(self):
        regex_op = api._get_regex_op("mysql://user:passw@localhost/sahara")
        self.assertEqual("REGEXP", regex_op)

        regex_op = api._get_regex_op("postgresql://localhost/sahara")
        self.assertEqual("~", regex_op)

        regex_op = api._get_regex_op("sqlite://user:passw@localhost/sahara")
        self.assertIsNone(regex_op)


class TestRegexFilter(base.SaharaWithDbTestCase):

    @mock.patch("sahara.db.sqlalchemy.api._get_regex_op")
    def test_regex_filter(self, get_regex_op):
        query = api.model_query(m.ClusterTemplate, context.ctx())

        regex_cols = ["name", "description", "plugin_name"]
        search_opts = {"name": "fred",
                       "hadoop_version": "2",
                       "bogus": "jack",
                       "plugin_name": "vanilla"}

        # Since regex_op is None remaining_opts should be a copy of search_opts
        get_regex_op.return_value = None
        query, remaining_opts = api.regex_filter(
            query, m.ClusterTemplate, regex_cols, search_opts)
        self.assertEqual(search_opts, remaining_opts)
        self.assertIsNot(search_opts, remaining_opts)

        # Since regex_cols is [] remaining_opts should be a copy of search_opts
        get_regex_op.return_value = "REGEXP"
        query, remaining_opts = api.regex_filter(
            query, m.ClusterTemplate, [], search_opts)
        self.assertEqual(search_opts, remaining_opts)
        self.assertIsNot(search_opts, remaining_opts)

        # Remaining should be search_opts with name and plugin_name removed
        # These are the only fields that are in regex_cols and also in
        # the model.
        get_regex_op.return_value = "REGEXP"
        query, remaining_opts = api.regex_filter(
            query, m.ClusterTemplate, regex_cols, search_opts)
        self.assertEqual({"hadoop_version": "2",
                          "bogus": "jack"}, remaining_opts)

        # bogus is not in the model so it should be left in remaining
        # even though regex_cols lists it
        regex_cols.append("bogus")
        query, remaining_opts = api.regex_filter(
            query, m.ClusterTemplate, regex_cols, search_opts)
        self.assertEqual({"hadoop_version": "2",
                          "bogus": "jack"}, remaining_opts)

        # name will not be removed because the value is not a string
        search_opts["name"] = 5
        query, remaining_opts = api.regex_filter(
            query, m.ClusterTemplate, regex_cols, search_opts)
        self.assertEqual({"hadoop_version": "2",
                          "bogus": "jack",
                          "name": 5}, remaining_opts)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import re

from oslo_serialization import jsonutils as json
from oslo_utils import uuidutils
import six
from six.moves.urllib import parse as urlparse

import sahara.exceptions as ex


@six.add_metaclass(abc.ABCMeta)
class BaseOozieClient(object):
    def __init__(self, url, oozie_server):
        self.job_url = url + "/v2/job/%s"
        self.jobs_url = url + "/v2/jobs"
        self.oozie_server = oozie_server
        self.port = urlparse.urlparse(url).port

    @abc.abstractmethod
    def add_job(self, job_config, job_execution):
        pass

    @abc.abstractmethod
    def manage_job(self, job_execution, action, job_id=None):
        pass

    @abc.abstractmethod
    def get_job_info(self, job_execution, job_id=None):
        pass

    def kill_job(self, job_execution):
        self.manage_job(job_execution, 'kill')

    def run_job(self, job_execution, job_id):
        self.manage_job(job_execution, 'start', job_id=job_id)


class OozieClient(BaseOozieClient):
    def add_job(self, job_config, job_execution):
        return self.post(
            job_execution, self.jobs_url, data=job_config, headers={
                "Content-Type": "application/xml;charset=UTF-8"})

    def manage_job(self, job_execution, action, job_id=None):
        job_id = job_id if job_id else job_execution.engine_job_id
        url = self.job_url % job_id + "?action=" + action
        self.put(job_execution, url)

    def get_job_info(self, job_execution, job_id=None):
        job_id = job_id if job_id else job_execution.engine_job_id
        url = self.job_url % job_id + "?show=info"
        return self.get(job_execution, url)

    def _get_http_session(self, info=None):
        return self.oozie_server.remote().get_http_client(self.port, info=info)

    def post(self, job_execution, url, data, headers):
        session = self._get_http_session(job_execution.extra.get('neutron'))
        resp = session.post(url, data=data, headers=headers)
        _check_status_code(resp, 201)
        return get_json(resp)['id']

    def put(self, job_execution, url):
        session = self._get_http_session(job_execution.extra.get('neutron'))
        resp = session.put(url)
        _check_status_code(resp, 200)

    def get(self, job_execution, url):
        session = self._get_http_session(job_execution.extra.get('neutron'))
        resp = session.get(url)
        _check_status_code(resp, 200)
        return get_json(resp)


class RemoteOozieClient(OozieClient):
    def __init__(self, url, oozie_server, hdfs_user):
        self.hdfs_user = hdfs_user
        self.oozie_url = url.replace(
            urlparse.urlparse(url).hostname, oozie_server.fqdn())
        super(RemoteOozieClient, self).__init__(url, oozie_server)

    def _oozie(self, cmd):
        return (
            "sudo su - -c 'oozie -Doozie.auth.token.cache=false "
            "{cmd} -oozie {oozie}' {user}".format(
                cmd=cmd, oozie=self.oozie_url, user=self.hdfs_user))

    def add_job(self, job_config, job_execution):
        with self.oozie_server.remote() as r:
            name = "/tmp/%s.xml" % uuidutils.generate_uuid()[:8]
            r.write_file_to(name, job_config)
            cmd = self._oozie("job -submit -config %s" % name)
            cmd += " | awk '{ print $2 }'"
            code, stdout = r.execute_command(cmd)
        stdout = stdout.strip()
        return stdout

    def manage_job(self, job_execution, action, job_id=None):
        job_id = job_id if job_id else job_execution.engine_job_id
        cmd = self._oozie("job -%s %s" % (action, job_id))
        with self.oozie_server.remote() as r:
            r.execute_command(cmd)

    def get_job_info(self, job_execution, job_id=None):
        job_id = job_id if job_id else job_execution.engine_job_id
        cmd = self._oozie("job -info %s" % job_id)
        cmd += " | grep Status | head -n 1 | awk '{ print $3 }'"
        with self.oozie_server.remote() as r:
            code, stdout = r.execute_command(cmd)
        return {'status': stdout.strip()}


def _check_status_code(resp, expected_code):
    if resp.status_code != expected_code:
        resp_text = resp.text
        # cleaning tomcat error message
        message = resp_text.split("<HR size=\"1\" noshade=\"noshade\">")[1]
        message = message.replace("</p><p>", "\n")
        message = re.sub('<[^<]+?>', ' ', message)
        raise ex.OozieException(message)


def get_json(response):
    """Provides backward compatibility for old versions of requests library."""

    json_field_or_function = getattr(response, 'json', None)
    if callable(json_field_or_function):
        return response.json()
    else:
        return json.loads(response.content)



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six


class Command(object):
    def __init__(self, option_values):
        for k, v in six.iteritems(option_values):
            setattr(self, k, v)


class Config(object):
    def __init__(self, option_values=None):
        self.command = Command(option_values or {})


class Logger(object):
    def __init__(self):
        self.clear_log()

    def warning(self, message):
        self.warnings.append(message)

    def info(self, message):
        self.infos.append(message)

    def debug(self, message):
        pass

    def clear_log(self):
        self.warnings = []
        self.infos = []


SAMPLE_NGT = {
    "plugin_name": "test_plugin",
    "flavor_id": "42",
    "tenant_id": "tenant_1",
    "hadoop_version": "test_version",
    "name": "ngt_test",
    "node_processes": ["p1", "p2"],
    "floating_ip_pool": None,
    "availability_zone": None,
    "node_configs": {
        "service_1": {
            "config_1": "value_1"
        },
        "service_2": {
            "config_1": "value_1"
        },
    },
    "is_default": True
}

SAMPLE_CLT = {
    "plugin_name": "test_plugin",
    "tenant_id": "tenant_1",
    "hadoop_version": "test_version",
    "name": "clt-test",
    "cluster_configs": {
        "service_1": {
            "config_1": "value_1"
        },
        "service_2": {
            "config_1": "value_1"
        }
    },
    "node_groups": [
        {
            "name": "ng_1",
            "flavor_id": "42",
            "node_processes": ["p1", "p2"],
            "count": 1,
            "floating_ip_pool": None,
            "security_groups": None,
            "availability_zone": None,
        }
    ],
    "is_default": True
}

SAMPLE_CLUSTER = {
    "name": "test_cluster",
    "plugin_name": "test_plugin",
    "hadoop_version": "test_version",
    "tenant_id": "tenant_1",
    "node_groups": [
        {
            "name": "ng_1",
            "node_group_template_id": "ng_1_id",
            "count": 1,
        }
    ]
}



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara.service import ntp_service as ntp
from sahara.tests.unit import base as test_base


class FakeRemote(object):
    def __init__(self, effects):
        self.effects = effects
        self.idx = 0

    def __enter__(self):
        return self

    def __exit__(self, *args):
        # validate number of executions
        if self.idx != len(self.effects):
            raise ValueError()

    def _get_effect(self):
        self.idx += 1
        return self.effects[self.idx - 1]

    def execute_command(self, cmd, run_as_root=False):
        effect = self._get_effect()
        if isinstance(effect, RuntimeError):
            raise effect
        return 0, effect

    def append_to_file(self, file, text, run_as_root=False):
        return self.execute_command(file, run_as_root)

    def prepend_to_file(self, file, text, run_as_root=False):
        return self.execute_command(file, run_as_root)

    def get_os_distrib(self):
        return self.execute_command('get_os_distrib')


class FakeInstance(object):
    def __init__(self, effects, id):
        self.id = id
        self.instance_name = id
        self.instance_id = id
        self.effects = effects

    def remote(self):
        return FakeRemote(self.effects)


class NTPServiceTest(test_base.SaharaTestCase):
    @mock.patch('sahara.service.ntp_service.LOG.warning')
    @mock.patch('sahara.service.ntp_service.conductor.cluster_get')
    def test_configuring_ntp_unable_to_configure(self, cl_get, logger):
        instance = FakeInstance(["ubuntu", RuntimeError()], "1")
        ng = mock.Mock(instances=[instance])
        cl_get.return_value = mock.Mock(
            node_groups=[ng], cluster_configs={})
        ntp.configure_ntp('1')
        self.assertEqual(
            [mock.call("Unable to configure NTP service")],
            logger.call_args_list)

    @mock.patch('sahara.service.ntp_service.LOG.info')
    @mock.patch('sahara.service.ntp_service.conductor.cluster_get')
    def test_configuring_success(self, cl_get, logger):
        instance = FakeInstance(
            ['centos', "cat", "batman", "vs", "superman", "boom"], "1")
        ng = mock.Mock(instances=[instance])
        cl_get.return_value = mock.Mock(node_groups=[ng], cluster_configs={})
        ntp.configure_ntp('1')
        self.assertEqual([mock.call("NTP successfully configured")],
                         logger.call_args_list)

    def test_retrieve_url(self):
        cl = mock.Mock(
            cluster_configs={'general': {"URL of NTP server": "batman.org"}})
        self.assertEqual("batman.org", ntp.retrieve_ntp_server_url(cl))
        self.override_config('default_ntp_server', "superman.org")
        cl = mock.Mock(cluster_configs={'general': {}})
        self.assertEqual("superman.org", ntp.retrieve_ntp_server_url(cl))

    @mock.patch('sahara.service.ntp_service.conductor.cluster_get')
    @mock.patch('sahara.service.ntp_service.retrieve_ntp_server_url')
    def test_is_ntp_enabled(self, ntp_url, cl_get):
        cl = mock.Mock(
            cluster_configs={'general': {"Enable NTP service": False}})
        cl_get.return_value = cl
        ntp.configure_ntp('1')
        self.assertEqual(0, ntp_url.call_count)



# Copyright (c) 2014 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_utils import uuidutils

from sahara.utils import files

# job execution status
JOB_STATUS_DONEWITHERROR = 'DONEWITHERROR'
JOB_STATUS_FAILED = 'FAILED'
JOB_STATUS_KILLED = 'KILLED'
JOB_STATUS_PENDING = 'PENDING'
JOB_STATUS_READYTORUN = 'READYTORUN'
JOB_STATUS_RUNNING = 'RUNNING'
JOB_STATUS_SUCCEEDED = 'SUCCEEDED'
JOB_STATUS_TOBEKILLED = 'TOBEKILLED'
JOB_STATUS_TOBESUSPENDED = 'TOBESUSPENDED'
JOB_STATUS_PREP = 'PREP'
JOB_STATUS_PREPSUSPENDED = 'PREPSUSPENDED'
JOB_STATUS_SUSPENDED = 'SUSPENDED'
JOB_STATUS_SUSPEND_FAILED = 'SUSPENDFAILED'
# statuses for suspended jobs
JOB_STATUSES_SUSPENDIBLE = [
    JOB_STATUS_PREP,
    JOB_STATUS_RUNNING
]
# statuses for terminated jobs
JOB_STATUSES_TERMINATED = [
    JOB_STATUS_DONEWITHERROR,
    JOB_STATUS_FAILED,
    JOB_STATUS_KILLED,
    JOB_STATUS_SUCCEEDED,
    JOB_STATUS_SUSPEND_FAILED
]
# job type separator character
JOB_TYPE_SEP = '.'
# job sub types available
JOB_SUBTYPE_STREAMING = 'Streaming'
JOB_SUBTYPE_NONE = ''
# job types available
JOB_TYPE_HIVE = 'Hive'
JOB_TYPE_JAVA = 'Java'
JOB_TYPE_MAPREDUCE = 'MapReduce'
JOB_TYPE_SPARK = 'Spark'
JOB_TYPE_STORM = 'Storm'
JOB_TYPE_PYLEUS = 'Storm.Pyleus'
JOB_TYPE_MAPREDUCE_STREAMING = (JOB_TYPE_MAPREDUCE + JOB_TYPE_SEP +
                                JOB_SUBTYPE_STREAMING)
JOB_TYPE_PIG = 'Pig'
JOB_TYPE_SHELL = 'Shell'

# job type groupings available
JOB_TYPES_ALL = [
    JOB_TYPE_HIVE,
    JOB_TYPE_JAVA,
    JOB_TYPE_MAPREDUCE,
    JOB_TYPE_MAPREDUCE_STREAMING,
    JOB_TYPE_PIG,
    JOB_TYPE_SHELL,
    JOB_TYPE_SPARK,
    JOB_TYPE_STORM,
    JOB_TYPE_PYLEUS
]

JOB_TYPES_ACCEPTABLE_CONFIGS = {
    JOB_TYPE_HIVE: {"configs", "params"},
    JOB_TYPE_PIG: {"configs", "params", "args"},
    JOB_TYPE_MAPREDUCE: {"configs"},
    JOB_TYPE_MAPREDUCE_STREAMING: {"configs"},
    JOB_TYPE_JAVA: {"configs", "args"},
    JOB_TYPE_SHELL: {"configs", "params", "args"},
    JOB_TYPE_SPARK: {"configs", "args"},
    JOB_TYPE_STORM: {"args"},
    JOB_TYPE_PYLEUS: {}
}

# job actions
JOB_ACTION_SUSPEND = 'suspend'
JOB_ACTION_CANCEL = 'cancel'

JOB_ACTION_TYPES_ACCEPTABLE = [
    JOB_ACTION_SUSPEND,
    JOB_ACTION_CANCEL
]

ADAPT_FOR_OOZIE = 'edp.java.adapt_for_oozie'
SPARK_DRIVER_CLASSPATH = 'edp.spark.driver.classpath'
ADAPT_SPARK_FOR_SWIFT = 'edp.spark.adapt_for_swift'


def split_job_type(job_type):
    '''Split a job type string into a type and subtype

    The split is done on the first '.'.  A subtype will
    always be returned, even if it is empty.
    '''
    type_info = job_type.split(JOB_TYPE_SEP, 1)
    if len(type_info) == 1:
        type_info.append('')
    return type_info


def compare_job_type(job_type, *args, **kwargs):
    '''Compare a job type against a list of job types

    :param job_type: The job type being compared
    :param *args: A list of types to compare against
    :param strict: Passed as a keyword arg. Default is False.
                   If strict is False, job_type will be compared
                   with and without its subtype indicator.
    :returns: True if job_type is present in the list, False otherwise
    '''
    strict = kwargs.get('strict', False)
    res = job_type in args
    if res or strict or JOB_TYPE_SEP not in job_type:
        return res

    jtype, jsubtype = split_job_type(job_type)
    return jtype in args


def get_hive_shared_conf_path(hdfs_user):
    return "/user/%s/conf/hive-site.xml" % hdfs_user


def is_adapt_for_oozie_enabled(configs):
    return configs.get(ADAPT_FOR_OOZIE, False)


def is_adapt_spark_for_swift_enabled(configs):
    return configs.get(ADAPT_SPARK_FOR_SWIFT, False)


def spark_driver_classpath(configs):
    # Return None in case when you need to use default value
    return configs.get(SPARK_DRIVER_CLASSPATH)


def get_builtin_binaries(job, configs):
    if job.type == JOB_TYPE_JAVA:
        if is_adapt_for_oozie_enabled(configs):
            path = 'service/edp/resources/edp-main-wrapper.jar'
            name = 'builtin-%s.jar' % uuidutils.generate_uuid()
            return [{'raw': files.get_file_binary(path),
                     'name': name}]
    return []



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import mock
import testtools

from sahara.utils import api
from sahara.utils import types
from sahara.utils import wsgi


class APIUtilsTest(testtools.TestCase):

    class FakeCluster(object):

        def to_dict(self):
            return {"id": 42, "name": "myFirstCluster"}

    page = types.Page([FakeCluster()])
    response = {"clusters":
                [
                    {
                        "id": 42,
                        "name": "myFirstCluster"
                    }
                ]
                }

    @mock.patch('flask.request')
    @mock.patch('flask.Response')
    def test_render_pagination(self, flask, request):
        serializer = wsgi.JSONDictSerializer()
        request.status_code = 200

        api.render(self.page, 'application/json', 200, name='clusters')
        body = serializer.serialize(self.response)
        flask.assert_called_with(
            response=body, status=200, mimetype='application/json')

        self.page.prev, self.page.next = 35, 49
        api.render(self.page, 'application/json', 200, name='clusters')
        paginate_response = copy.copy(self.response)
        paginate_response["markers"] = \
            {"prev": 35, "next": 49}

        body = serializer.serialize(paginate_response)
        flask.assert_called_with(
            response=body, status=200, mimetype='application/json')

        self.page.prev, self.page.next = 7, None
        api.render(self.page, 'application/json', 200, name='clusters')
        paginate_response = copy.copy(self.response)
        paginate_response["markers"] = {"prev": 7, "next": None}

        body = serializer.serialize(paginate_response)
        flask.assert_called_with(
            response=body, status=200, mimetype='application/json')

        self.page.prev, self.page.next = None, 14
        api.render(self.page, 'application/json', 200, name='clusters')

        paginate_response = copy.copy(self.response)
        paginate_response["markers"] = {"prev": None, "next": 14}

        body = serializer.serialize(paginate_response)
        flask.assert_called_with(
            response=body, status=200, mimetype='application/json')

        self.page.prev, self.page.next = None, 11
        api.render(self.page, 'application/json', 200, name='clusters')

        paginate_response = copy.copy(self.response)
        paginate_response["markers"] = \
            {"prev": None, "next": 11}

        body = serializer.serialize(paginate_response)
        flask.assert_called_with(
            response=body, status=200, mimetype='application/json')

        self.page.prev, self.page.next = None, 11
        api.render(self.page, 'application/json', 200, name='clusters')



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""health-check

Revision ID: 029
Revises: 028
Create Date: 2016-01-26 16:11:46.008367

"""

# revision identifiers, used by Alembic.
revision = '030'
down_revision = '029'

from alembic import op
import sqlalchemy as sa


MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    op.create_table(
        'cluster_verifications',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('cluster_id', sa.String(length=36), nullable=True),
        sa.Column('status', sa.String(length=15), nullable=True),
        sa.ForeignKeyConstraint(['cluster_id'], ['clusters.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id', 'cluster_id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET)
    op.create_table(
        'cluster_health_checks',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('verification_id', sa.String(length=36), nullable=True),
        sa.Column('status', sa.String(length=15), nullable=True),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('name', sa.String(length=80), nullable=True),
        sa.ForeignKeyConstraint(
            ['verification_id'], ['cluster_verifications.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id', 'verification_id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET)



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.api import acl
from sahara.service.api.v2 import data_sources as api
from sahara.service import validation as v
from sahara.service.validations.edp import data_source as v_d_s
from sahara.service.validations.edp import data_source_schema as v_d_s_schema
import sahara.utils.api as u


rest = u.RestV2('data-sources', __name__)


@rest.get('/data-sources')
@acl.enforce("data-processing:data-source:list")
@v.check_exists(api.get_data_source, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_data_sources)
@v.validate_request_params(['type'])
def data_sources_list():
    result = api.get_data_sources(**u.get_request_args().to_dict())
    for ds in result:
        u._replace_tenant_id_project_id(ds)
    return u.render(res=result, name='data_sources')


@rest.post('/data-sources')
@acl.enforce("data-processing:data-source:register")
@v.validate(v_d_s_schema.DATA_SOURCE_SCHEMA, v_d_s.check_data_source_create)
@v.validate_request_params([])
def data_source_register(data):
    result = api.register_data_source(data).to_wrapped_dict()
    u._replace_tenant_id_project_id(result['data_source'])
    return u.render(result)


@rest.get('/data-sources/<data_source_id>')
@acl.enforce("data-processing:data-source:get")
@v.check_exists(api.get_data_source, 'data_source_id')
@v.validate_request_params([])
def data_source_get(data_source_id):
    result = api.get_data_source(data_source_id).to_wrapped_dict()
    u._replace_tenant_id_project_id(result['data_source'])
    return u.render(result)


@rest.delete('/data-sources/<data_source_id>')
@acl.enforce("data-processing:data-source:delete")
@v.check_exists(api.get_data_source, 'data_source_id')
@v.validate_request_params([])
def data_source_delete(data_source_id):
    api.delete_data_source(data_source_id)
    return u.render()


@rest.patch('/data-sources/<data_source_id>')
@acl.enforce("data-processing:data-source:update")
@v.check_exists(api.get_data_source, 'data_source_id')
@v.validate(v_d_s_schema.DATA_SOURCE_UPDATE_SCHEMA)
@v.validate_request_params([])
def data_source_update(data_source_id, data):
    result = api.data_source_update(data_source_id, data).to_wrapped_dict()
    u._replace_tenant_id_project_id(result['data_source'])
    return u.render(result)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flask
from oslo_config import cfg
import six
from werkzeug import exceptions as werkzeug_exceptions

from sahara.api import v10 as api_v10
from sahara.api import v11 as api_v11
from sahara.api import v2 as api_v2
from sahara import context
from sahara.utils import api as api_utils


CONF = cfg.CONF


def build_app():
    """App builder (wsgi).

    Entry point for Sahara REST API server
    """
    app = flask.Flask('sahara.api')

    @app.teardown_request
    def teardown_request(_ex=None):
        context.set_ctx(None)

    app.register_blueprint(api_v10.rest, url_prefix='/v1.0')
    app.register_blueprint(api_v10.rest, url_prefix='/v1.1')
    app.register_blueprint(api_v11.rest, url_prefix='/v1.1')

    def make_json_error(ex):
        status_code = (ex.code
                       if isinstance(ex, werkzeug_exceptions.HTTPException)
                       else 500)
        description = (ex.description
                       if isinstance(ex, werkzeug_exceptions.HTTPException)
                       else str(ex))
        return api_utils.render({'error': status_code,
                                 'error_message': description},
                                status=status_code)

    for code in six.iterkeys(werkzeug_exceptions.default_exceptions):
        app.register_error_handler(code, make_json_error)

    return app


def build_v2_app():
    """App builder (wsgi).

    Entry point for Experimental V2 Sahara REST API server
    """
    app = build_app()

    api_v2.register_blueprints(app, url_prefix='/v2')

    return app


class Router(object):
    def __call__(self, environ, response):
        return self.app(environ, response)

    @classmethod
    def factory(cls, global_config, **local_config):
        cls.app = build_app()
        return cls(**local_config)


class RouterV2(object):
    def __call__(self, environ, response):
        return self.app(environ, response)

    @classmethod
    def factory(cls, global_config, **local_config):
        cls.app = build_v2_app()
        return cls(**local_config)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import six


class BaseResource(object):
    __resource_name__ = 'base'
    __filter_cols__ = []

    @property
    def dict(self):
        return self.to_dict()

    @property
    def wrapped_dict(self):
        return {self.__resource_name__: self.dict}

    @property
    def __all_filter_cols__(self):
        cls = self.__class__
        if not hasattr(cls, '__mro_filter_cols__'):
            filter_cols = []
            for base_cls in inspect.getmro(cls):
                filter_cols += getattr(base_cls, '__filter_cols__', [])
            cls.__mro_filter_cols__ = set(filter_cols)
        return cls.__mro_filter_cols__

    def _filter_field(self, k):
        return k == '_sa_instance_state' or k in self.__all_filter_cols__

    def to_dict(self):
        dictionary = self.__dict__.copy()
        return {k: v for k, v in six.iteritems(dictionary)
                if not self._filter_field(k)}

    def as_resource(self):
        return Resource(self.__resource_name__, self.to_dict())


class Resource(BaseResource):
    def __init__(self, _name, _info):
        self._name = _name
        self.__resource_name__ = _name
        self._info = _info

    def __getattr__(self, k):
        if k not in self.__dict__:
            return self._info.get(k)
        return self.__dict__[k]

    def __repr__(self):
        return '<%s %s>' % (self._name, self._info)

    def __eq__(self, other):
        return self._name == other._name and self._info == other._info

    def to_dict(self):
        return self._info.copy()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bisect
import hashlib

from oslo_config import cfg
from oslo_log import log
from oslo_utils import uuidutils
from tooz import coordination

LOG = log.getLogger(__name__)

coordinator_opts = [
    cfg.IntOpt('coordinator_heartbeat_interval',
               default=1,
               help='Interval size between heartbeat execution in seconds. '
                    'Heartbeats are executed to make sure that connection to '
                    'the coordination server is active.'),
    cfg.IntOpt('hash_ring_replicas_count',
               default=40,
               help='Number of points that belongs to each member on a hash '
                    'ring. The larger number leads to a better distribution.')
]

CONF = cfg.CONF
CONF.register_opts(coordinator_opts)


class Coordinator(object):
    def __init__(self, backend_url):
        self.coordinator = None
        self.member_id = uuidutils.generate_uuid()

        if backend_url:
            try:
                self.coordinator = coordination.get_coordinator(
                    backend_url, self.member_id)
                self.coordinator.start()
                LOG.info('Coordination backend loaded successfully.')
            except coordination.ToozError:
                LOG.error('Error connecting to coordination backend.')
                raise

    def is_started(self):
        if self.coordinator:
            return self.coordinator.is_started
        return False

    def heartbeat(self):
        if self.coordinator:
            self.coordinator.heartbeat()

    def join_group(self, group_id):
        if self.coordinator:
            try:
                self.coordinator.join_group(group_id).get()
            except coordination.GroupNotCreated:
                try:
                    self.coordinator.create_group(group_id).get()
                except coordination.GroupAlreadyExist:
                    pass
                self.coordinator.join_group(group_id).get()

    def get_members(self, group_id):
        if self.coordinator:
            for i in range(2):
                try:
                    members = self.coordinator.get_members(group_id).get()
                    if self.member_id in members:
                        return members
                    self.join_group(group_id)
                except coordination.GroupNotCreated:
                    self.join_group(group_id)
                except coordination.ToozError as e:
                    LOG.error("Couldn't get members of {group} group. "
                              "Reason: {ex}".format(
                                  group=group_id, ex=str(e)))
        return []


class HashRing(Coordinator):
    def __init__(self, backend_url, group_id):
        self.group_id = group_id
        self.replicas = CONF.hash_ring_replicas_count
        super(HashRing, self).__init__(backend_url)
        self.join_group(group_id)

    @staticmethod
    def _hash(key):
        return int(
            hashlib.md5(str(key).encode('utf-8')).hexdigest(), 16)  # nosec

    def _build_ring(self):
        ring = {}
        members = self.get_members(self.group_id)
        for member in members:
            for r in range(self.replicas):
                hashed_key = self._hash('%s:%s' % (member, r))
                ring[hashed_key] = member

        return ring, sorted(ring.keys())

    def _check_object(self, object, ring, sorted_keys):
        """Checks if this object belongs to this member or not"""
        hashed_key = self._hash(object.id)
        position = bisect.bisect(sorted_keys, hashed_key)
        position = position if position < len(sorted_keys) else 0
        return ring[sorted_keys[position]] == self.member_id

    def get_subset(self, objects):
        """Returns subset that belongs to this member"""
        if self.coordinator:
            ring, keys = self._build_ring()
            if ring:
                return [obj for obj in objects if self._check_object(
                    obj, ring, keys)]
            return []
        return objects



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara.service import coordinator
from sahara.tests.unit import base


class TestCoordinator(base.SaharaTestCase):
    def test_coord_without_backend(self):
        coord = coordinator.Coordinator('')
        self.assertIsNone(coord.coordinator)

    @mock.patch('tooz.coordination.get_coordinator')
    def test_coord_with_backend(self, get_coord):
        MockCoord = mock.Mock()
        MockCoord.start.return_value = mock.Mock()
        get_coord.return_value = MockCoord

        coord = coordinator.Coordinator('kazoo://1.2.3.4:2181')
        self.assertEqual(MockCoord, coord.coordinator)
        MockCoord.start.assert_called_once_with()


class TestHashRing(base.SaharaTestCase):
    def setUp(self):
        super(TestHashRing, self).setUp()
        self.override_config('hash_ring_replicas_count', 1)

    @mock.patch('tooz.coordination.get_coordinator', return_value=mock.Mock())
    def _init_hr(self, get_coord):
        self.hr = coordinator.HashRing('kazoo://1.2.3.4:2181', 'group')
        self.hr.get_members = mock.Mock(return_value=['id1', 'id2', 'id3'])
        self.hr.member_id = 'id2'
        self.hr._hash = mock.Mock(side_effect=[1, 10, 20, 5, 13, 25])

    def test_get_subset_without_backend(self):
        hr = coordinator.HashRing('', 'group')
        objects = [mock.Mock(id=1), mock.Mock(id=2)]
        # all objects will be managed by this engine if coordinator backend
        # is not provided
        self.assertEqual(objects, hr.get_subset(objects))

    def test_build_ring(self):
        # check hash ring with one replica
        self._init_hr()
        hr, keys = self.hr._build_ring()
        self.assertEqual({1: 'id1', 10: 'id2', 20: 'id3'}, hr)
        self.assertEqual([1, 10, 20], keys)

        # check hash ring with two replicas
        self.override_config('hash_ring_replicas_count', 2)
        self._init_hr()
        hr, keys = self.hr._build_ring()
        self.assertEqual(
            {1: 'id1', 5: 'id2', 10: 'id1', 13: 'id3', 20: 'id2', 25: 'id3'},
            hr)
        self.assertEqual([1, 5, 10, 13, 20, 25], keys)

    def test_check_object(self):
        self._init_hr()
        ring, keys = self.hr._build_ring()

        # this object will be managed by this engine
        self.assertTrue(
            self.hr._check_object(mock.Mock(id='123'), ring, keys))

        # this object will not be managed by this engine
        self.assertFalse(
            self.hr._check_object(mock.Mock(id='321'), ring, keys))

        # this object will not be managed by this engine
        self.assertFalse(
            self.hr._check_object(mock.Mock(id='213'), ring, keys))

    def test_get_subset_with_backend(self):
        self._init_hr()
        objects = [mock.Mock(id=123), mock.Mock(id=321), mock.Mock(id=213)]

        # only first object will be managed by this engine
        self.assertEqual([objects[0]], self.hr.get_subset(objects))



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from sahara import context
from sahara.db.templates import utils
from sahara.tests.unit.conductor import base
from sahara.tests.unit.db.templates import common as c


class FakeNGT(object):
    def __init__(self, id):
        self.node_group_template_id = id


class FakeCluster(object):
    def __init__(self, name, node_groups=None, cluster_template_id=None):
        self.name = name
        self.node_groups = node_groups or []
        self.cluster_template_id = cluster_template_id


class TemplateUtilsTestCase(base.ConductorManagerTestCase):

    def test_substitute_ng_ids(self):
        cl = {"node_groups":
              [{"name": "worker",
                "node_group_template_id": "{vanilla-worker}",
                "count": 3},

               {"name": "master",
                "node_group_template_id": "{vanilla-master}",
                "count": 1},

               {"name": "secondary-name",
                "node_group_template_id": "some_id"}]}

        ng_dict = {"vanilla-worker": 1,
                   "vanilla-master": 2}

        utils.substitute_ng_ids(cl, ng_dict)
        self.assertEqual("1", cl["node_groups"][0]["node_group_template_id"])
        self.assertEqual("2", cl["node_groups"][1]["node_group_template_id"])
        self.assertEqual("some_id",
                         cl["node_groups"][2]["node_group_template_id"])

    def test_check_plugin_version(self):

        template = {"plugin_name": "vanilla",
                    "hadoop_version": "2.7.1"}

        self.assertTrue(utils.check_plugin_version(template, None))
        self.assertTrue(utils.check_plugin_version(template, ["2.7.1"]))
        self.assertTrue(utils.check_plugin_version(template,
                                                   ["vanilla.2.7.1"]))
        self.assertFalse(utils.check_plugin_version(template, ["1.2.1"]))

    def test_check_plugin_name_and_version(self):

        template = {"plugin_name": "vanilla",
                    "hadoop_version": "2.7.1"}

        self.assertTrue(utils.check_plugin_name_and_version(
            template, None, ["2.7.1"]))
        self.assertTrue(utils.check_plugin_name_and_version(
            template, ["vanilla"], None))
        self.assertTrue(utils.check_plugin_name_and_version(
            template, ["vanilla"], ["2.7.1"]))
        self.assertTrue(utils.check_plugin_name_and_version(
            template, ["vanilla"], ["vanilla.2.7.1"]))
        self.assertFalse(utils.check_plugin_name_and_version(
            template, ["hdp"], ["2.7.1"]))

    def test_check_node_group_template_usage(self):

        ng1 = FakeNGT(1)
        ng2 = FakeNGT(2)

        cluster = FakeCluster("cluster", [ng1])
        template = FakeCluster("template", [ng2])

        cluster_users, template_users = utils.check_node_group_template_usage(
            1, [cluster], [template])
        self.assertEqual([cluster.name], cluster_users)
        self.assertEqual([], template_users)

        cluster_users, template_users = utils.check_node_group_template_usage(
            2, [cluster], [template])
        self.assertEqual([], cluster_users)
        self.assertEqual([template.name], template_users)

    def test_check_cluster_template_usage(self):
        cluster = FakeCluster("cluster", cluster_template_id=1)

        cluster_users = utils.check_cluster_template_usage(1, [cluster])
        self.assertEqual([cluster.name], cluster_users)

    def test_find_node_group_template_by_name(self):
        ctx = context.ctx()
        t = self.api.node_group_template_create(ctx, c.SAMPLE_NGT)

        found = utils.find_node_group_template_by_name(ctx,
                                                       c.SAMPLE_NGT["name"])
        self.assertEqual(t["id"], found["id"])

        found = utils.find_node_group_template_by_name(ctx, "fred")
        self.assertIsNone(found)

    def test_find_cluster_template_by_name(self):
        ctx = context.ctx()
        t = self.api.cluster_template_create(ctx, c.SAMPLE_CLT)

        found = utils.find_cluster_template_by_name(ctx, c.SAMPLE_CLT["name"])
        self.assertEqual(t["id"], found["id"])

        found = utils.find_cluster_template_by_name(ctx, "fred")
        self.assertIsNone(found)

    def test_value_diff(self):

        current = {"cat": "meow",
                   "dog": "woof",
                   "horse": ["neigh", "whinny"]}

        new_values = {"dog": "bark",
                      "horse": "snort"}

        original = copy.deepcopy(current)
        backup = utils.value_diff(current, new_values)

        self.assertEqual({"dog": "woof",
                          "horse": ["neigh", "whinny"]}, backup)

        # current is unchanged
        self.assertEqual(original, current)



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara.tests.unit import base
from sahara.utils.openstack import neutron as neutron_client


class NeutronClientTest(base.SaharaTestCase):
    @mock.patch("sahara.utils.openstack.keystone.token_auth")
    @mock.patch("neutronclient.neutron.client.Client")
    def test_get_router(self, patched, token_auth):
        patched.side_effect = _test_get_neutron_client
        neutron = neutron_client.NeutronClient(
            '33b47310-b7a8-4559-bf95-45ba669a448e', None, None)
        self.assertEqual('6c4d4e32-3667-4cd4-84ea-4cc1e98d18be',
                         neutron.get_router())


def _test_get_neutron_client(api_version, *args, **kwargs):
    return FakeNeutronClient()


class FakeNeutronClient(object):
    def list_routers(self):
        return {"routers": [{"status": "ACTIVE", "external_gateway_info": {
            "network_id": "61f95d3f-495e-4409-8c29-0b806283c81e"},
            "name": "router1", "admin_state_up": True,
            "tenant_id": "903809ded3434f8d89948ee71ca9f5bb",
            "routes": [],
            "id": "6c4d4e32-3667-4cd4-84ea-4cc1e98d18be"}]}

    def list_ports(self, device_id=None):
        return {"ports": [
            {"status": "ACTIVE", "name": "", "admin_state_up": True,
             "network_id": "33b47310-b7a8-4559-bf95-45ba669a448e",
             "tenant_id": "903809ded3434f8d89948ee71ca9f5bb",
             "binding:vif_type": "ovs", "device_owner": "compute:None",
             "binding:capabilities": {"port_filter": True},
             "mac_address": "fa:16:3e:69:25:1c", "fixed_ips": [
                {"subnet_id": "bfa9d0a1-9efb-4bff-bd2b-c103c053560f",
                 "ip_address": "10.0.0.8"}],
             "id": "0f3df685-bc55-4314-9b76-835e1767b78f",
             "security_groups": ["f9fee2a2-bb0b-44e4-8092-93a43dc45cda"],
             "device_id": "c2129c18-6707-4f07-94cf-00b2fef8eea7"},
            {"status": "ACTIVE", "name": "", "admin_state_up": True,
             "network_id": "33b47310-b7a8-4559-bf95-45ba669a448e",
             "tenant_id": "903809ded3434f8d89948ee71ca9f5bb",
             "binding:vif_type": "ovs",
             "device_owner": "network:router_interface",
             "binding:capabilities": {"port_filter": True},
             "mac_address": "fa:16:3e:c5:b0:cb", "fixed_ips": [
                 {"subnet_id": "bfa9d0a1-9efb-4bff-bd2b-c103c053560f",
                  "ip_address": "10.0.0.1"}],
             "id": "27193ae1-142a-436c-ab41-c77b1df032a1",
             "security_groups": [],
             "device_id": "6c4d4e32-3667-4cd4-84ea-4cc1e98d18be"}]}



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.api import acl
from sahara.service.api.v2 import cluster_templates as api
from sahara.service import validation as v
from sahara.service.validations import cluster_template_schema as ct_schema
from sahara.service.validations import cluster_templates as v_ct
import sahara.utils.api as u


rest = u.RestV2('cluster-templates', __name__)


@rest.get('/cluster-templates')
@acl.enforce("data-processing:cluster-template:list")
@v.check_exists(api.get_cluster_template, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_cluster_templates)
@v.validate_request_params(['plugin_name', 'plugin_version', 'name'])
def cluster_templates_list():
    request_args = u.get_request_args().to_dict()
    if 'plugin_version' in request_args:
        request_args['hadoop_version'] = request_args['plugin_version']
        del request_args['plugin_version']
    result = api.get_cluster_templates(**request_args)
    for ct in result:
        u._replace_hadoop_version_plugin_version(ct)
        u._replace_tenant_id_project_id(ct)
    return u.render(res=result, name='cluster_templates')


@rest.post('/cluster-templates')
@acl.enforce("data-processing:cluster-template:create")
@v.validate(ct_schema.CLUSTER_TEMPLATE_SCHEMA_V2,
            v_ct.check_cluster_template_create)
@v.validate_request_params([])
def cluster_templates_create(data):
    # renaming hadoop_version -> plugin_version
    # this can be removed once APIv1 is deprecated
    data['hadoop_version'] = data['plugin_version']
    del data['plugin_version']
    result = api.create_cluster_template(data).to_wrapped_dict()
    u._replace_hadoop_version_plugin_version(result['cluster_template'])
    u._replace_tenant_id_project_id(result['cluster_template'])
    return u.render(result)


@rest.get('/cluster-templates/<cluster_template_id>')
@acl.enforce("data-processing:cluster-template:get")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
@v.validate_request_params([])
def cluster_templates_get(cluster_template_id):
    result = u.to_wrapped_dict_no_render(
        api.get_cluster_template, cluster_template_id)
    u._replace_hadoop_version_plugin_version(result['cluster_template'])
    u._replace_tenant_id_project_id(result['cluster_template'])
    return u.render(result)


@rest.patch('/cluster-templates/<cluster_template_id>')
@acl.enforce("data-processing:cluster-template:update")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
@v.validate(ct_schema.CLUSTER_TEMPLATE_UPDATE_SCHEMA_V2,
            v_ct.check_cluster_template_update)
@v.validate_request_params([])
def cluster_templates_update(cluster_template_id, data):
    if data.get('plugin_version', None):
        data['hadoop_version'] = data['plugin_version']
        del data['plugin_version']
    result = u.to_wrapped_dict_no_render(
        api.update_cluster_template, cluster_template_id, data)
    u._replace_hadoop_version_plugin_version(result['cluster_template'])
    u._replace_tenant_id_project_id(result['cluster_template'])
    return u.render(result)


@rest.delete('/cluster-templates/<cluster_template_id>')
@acl.enforce("data-processing:cluster-template:delete")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
@v.validate(None, v_ct.check_cluster_template_usage)
@v.validate_request_params([])
def cluster_templates_delete(cluster_template_id):
    api.terminate_cluster_template(cluster_template_id)
    return u.render()


def _cluster_template_export_helper(template):
    template.pop('id')
    template.pop('updated_at')
    template.pop('created_at')
    template.pop('project_id')
    template.pop('is_default')
    template['default_image_id'] = '{default_image_id}'
    template['node_groups'] = '{node_groups}'


@rest.get('/cluster-templates/<cluster_template_id>/export')
@acl.enforce("data-processing:cluster-template:get")
@v.check_exists(api.get_cluster_template, 'cluster_template_id')
@v.validate_request_params([])
def cluster_template_export(cluster_template_id):
    content = u.to_wrapped_dict_no_render(
        api.export_cluster_template, cluster_template_id)
    u._replace_hadoop_version_plugin_version(content['cluster_template'])
    u._replace_tenant_id_project_id(content['cluster_template'])
    _cluster_template_export_helper(content['cluster_template'])
    res = u.render(content)
    res.headers.add('Content-Disposition', 'attachment',
                    filename='cluster_template.json')
    return res



# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Only (de)serialization utils hasn't been removed to decrease requirements
# number.

"""Utility methods for working with WSGI servers."""

import datetime

from oslo_serialization import jsonutils
import six

from sahara import exceptions
from sahara.i18n import _


class ActionDispatcher(object):
    """Maps method name to local methods through action name."""

    def dispatch(self, *args, **kwargs):
        """Find and call local method."""
        action = kwargs.pop('action', 'default')
        action_method = getattr(self, str(action), self.default)
        return action_method(*args, **kwargs)

    def default(self, data):
        raise NotImplementedError()


class DictSerializer(ActionDispatcher):
    """Default request body serialization."""

    def serialize(self, data, action='default'):
        return self.dispatch(data, action=action)

    def default(self, data):
        return ""


class JSONDictSerializer(DictSerializer):
    """Default JSON request body serialization."""

    def default(self, data):
        def sanitizer(obj):
            if isinstance(obj, datetime.datetime):
                _dtime = obj - datetime.timedelta(microseconds=obj.microsecond)
                return _dtime.isoformat()
            return six.text_type(obj)
        return jsonutils.dumps(data, default=sanitizer)


class TextDeserializer(ActionDispatcher):
    """Default request body deserialization."""

    def deserialize(self, datastring, action='default'):
        return self.dispatch(datastring, action=action)

    def default(self, datastring):
        return {}


class JSONDeserializer(TextDeserializer):

    def _from_json(self, datastring):
        try:
            return jsonutils.loads(datastring)
        except ValueError:
            msg = _("cannot understand JSON")
            raise exceptions.MalformedRequestBody(msg)

    def default(self, datastring):
        return {'body': self._from_json(datastring)}



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.api import acl
from sahara.service.api.v2 import job_templates as api
from sahara.service import validation as v
from sahara.service.validations.edp import job as v_j
from sahara.service.validations.edp import job_schema as v_j_schema
import sahara.utils.api as u


rest = u.RestV2('job-templates', __name__)


def _replace_tenant_id_project_id_job_binary(jb_list):
    for jb_obj in jb_list:
        dict.update(jb_obj, {'project_id': jb_obj['tenant_id']})
        dict.pop(jb_obj, 'tenant_id')


@rest.get('/job-templates')
@acl.enforce("data-processing:job-template:list")
@v.check_exists(api.get_job_template, 'marker')
@v.validate(None, v.validate_pagination_limit,
            v.validate_sorting_jobs)
@v.validate_request_params(['type', 'name'])
def job_templates_list():
    result = api.get_job_templates(**u.get_request_args().to_dict())
    for jt in result:
        u._replace_tenant_id_project_id(jt)
        _replace_tenant_id_project_id_job_binary(jt['mains'])
        _replace_tenant_id_project_id_job_binary(jt['libs'])
    return u.render(res=result, name='job_templates')


@rest.post('/job-templates')
@acl.enforce("data-processing:job-template:create")
@v.validate(v_j_schema.JOB_SCHEMA, v_j.check_mains_libs, v_j.check_interface)
@v.validate_request_params([])
def job_templates_create(data):
    result = {'job_template': api.create_job_template(data).to_dict()}
    u._replace_tenant_id_project_id(result['job_template'])
    _replace_tenant_id_project_id_job_binary(result['job_template']['mains'])
    _replace_tenant_id_project_id_job_binary(result['job_template']['libs'])
    return u.render(result)


@rest.get('/job-templates/<job_templates_id>')
@acl.enforce("data-processing:job-template:get")
@v.check_exists(api.get_job_template, id='job_templates_id')
@v.validate_request_params([])
def job_templates_get(job_templates_id):
    result = {'job_template': api.get_job_template(
        job_templates_id).to_dict()}
    u._replace_tenant_id_project_id(result['job_template'])
    _replace_tenant_id_project_id_job_binary(result['job_template']['mains'])
    _replace_tenant_id_project_id_job_binary(result['job_template']['libs'])
    return u.render(result)


@rest.patch('/job-templates/<job_templates_id>')
@acl.enforce("data-processing:job-template:update")
@v.check_exists(api.get_job_template, id='job_templates_id')
@v.validate(v_j_schema.JOB_UPDATE_SCHEMA)
@v.validate_request_params([])
def job_templates_update(job_templates_id, data):
    result = {'job_template': api.update_job_template(
        job_templates_id, data).to_dict()}
    u._replace_tenant_id_project_id(result['job_template'])
    _replace_tenant_id_project_id_job_binary(result['job_template']['mains'])
    _replace_tenant_id_project_id_job_binary(result['job_template']['libs'])
    return u.render(result)


@rest.delete('/job-templates/<job_templates_id>')
@acl.enforce("data-processing:job-template:delete")
@v.check_exists(api.get_job_template, id='job_templates_id')
@v.validate_request_params([])
def job_templates_delete(job_templates_id):
    api.delete_job_template(job_templates_id)
    return u.render()


@rest.get('/job-templates/config-hints/<job_type>')
@acl.enforce("data-processing:job-template:get-config-hints")
@v.check_exists(api.get_job_config_hints, job_type='job_type')
@v.validate_request_params([])
def job_config_hints_get(job_type):
    return u.render(api.get_job_config_hints(job_type))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sahara import context
import sahara.exceptions as ex
from sahara.i18n import _
import sahara.plugins.base as plugin_base
from sahara.service import api as service_api
from sahara.service.api import v10 as api
from sahara.service.validations import acl
import sahara.service.validations.base as b
from sahara.utils import cluster as c_u


def check_cluster_scaling(data, cluster_id, **kwargs):
    ctx = context.current()
    cluster = api.get_cluster(id=cluster_id)

    if cluster is None:
        raise ex.NotFoundException(
            {'id': cluster_id}, _('Object with %s not found'))

    b.check_plugin_labels(
        cluster.plugin_name, cluster.hadoop_version)

    acl.check_tenant_for_update(ctx, cluster)
    acl.check_protected_from_update(cluster, data)

    cluster_engine = cluster.sahara_info.get(
        'infrastructure_engine') if cluster.sahara_info else None

    engine_type_and_version = service_api.OPS.get_engine_type_and_version()
    if (not cluster_engine and
            not engine_type_and_version.startswith('direct')):
        raise ex.InvalidReferenceException(
            _("Cluster created before Juno release "
              "can't be scaled with %(engine)s engine") %
            {"engine": engine_type_and_version})

    if (cluster.sahara_info and
            cluster_engine != engine_type_and_version):
        raise ex.InvalidReferenceException(
            _("Cluster created with %(old_engine)s infrastructure engine "
              "can't be scaled with %(new_engine)s engine") %
            {"old_engine": cluster.sahara_info.get('infrastructure_engine'),
             "new_engine": engine_type_and_version})

    if not (plugin_base.PLUGINS.is_plugin_implements(cluster.plugin_name,
                                                     'scale_cluster') and (
            plugin_base.PLUGINS.is_plugin_implements(cluster.plugin_name,
                                                     'decommission_nodes'))):
        raise ex.InvalidReferenceException(
            _("Requested plugin '%s' doesn't support cluster scaling feature")
            % cluster.plugin_name)

    if cluster.status != c_u.CLUSTER_STATUS_ACTIVE:
        raise ex.InvalidReferenceException(
            _("Cluster cannot be scaled not in 'Active' status. "
              "Cluster status: %s") % cluster.status)

    if cluster.user_keypair_id:
        b.check_keypair_exists(cluster.user_keypair_id)

    if cluster.default_image_id:
        b.check_image_registered(cluster.default_image_id)

    if data.get("resize_node_groups"):
        b.check_resize(cluster, data['resize_node_groups'])

    if data.get("add_node_groups"):
        b.check_add_node_groups(cluster, data['add_node_groups'])
        b.check_cluster_hostnames_lengths(cluster.name,
                                          data['add_node_groups'])



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

import sahara.exceptions as ex
from sahara.service.castellan import config as castellan
from sahara.service.edp.job_binaries.swift.implementation import SwiftType
from sahara.tests.unit import base


class TestSwiftType(base.SaharaTestCase):

    def setUp(self):
        super(TestSwiftType, self).setUp()
        castellan.validate_config()
        self.i_s = SwiftType()

    def test_validate_job_location_format(self):
        self.assertFalse(self.i_s.validate_job_location_format(''))
        self.assertFalse(self.i_s.validate_job_location_format('swift://'))
        self.assertTrue(self.i_s.
                        validate_job_location_format('swift://123'))

    def test_validate(self):
        data = {}
        self.i_s.validate(data, job_binary_id='id')

        with testtools.ExpectedException(ex.BadJobBinaryException):
            self.i_s.validate(data)

        self.override_config('use_domain_for_proxy_users', True)
        self.i_s.validate(data)

        self.override_config('use_domain_for_proxy_users', False)
        data = {
            "extra": {
                "user": "user",
                "password": "pass"
            }
        }
        self.i_s.validate(data)

    @mock.patch('sahara.service.edp.job_binaries.'
                'swift.implementation.SwiftType.get_raw_data')
    def test_copy_binary_to_cluster(self, get_raw_data):
        remote = mock.Mock()
        job_binary = mock.Mock()
        job_binary.name = 'test'
        job_binary.url = 'swift://somebinary'
        get_raw_data.return_value = 'test'

        res = self.i_s.copy_binary_to_cluster(job_binary,
                                              remote=remote)

        self.assertEqual('/tmp/test', res)
        remote.write_file_to.assert_called_with(
            '/tmp/test',
            'test')

    def test__get_raw_data(self):
        client_instance = mock.Mock()
        client_instance.head_object = mock.Mock()
        client_instance.get_object = mock.Mock()

        job_binary = mock.Mock()
        job_binary.url = 'swift://container/object'

        # an object that is too large should raise an exception
        header = {'content-length': '2048'}
        client_instance.head_object.return_value = header
        self.override_config('job_binary_max_KB', 1)
        self.assertRaises(ex.DataTooBigException,
                          self.i_s._get_raw_data,
                          job_binary, client_instance)
        client_instance.head_object.assert_called_once_with('container',
                                                            'object')

        # valid return
        header = {'content-length': '4'}
        body = 'data'
        client_instance.head_object.return_value = header
        client_instance.get_object.return_value = (header, body)
        self.assertEqual(body, self.i_s._get_raw_data(job_binary,
                         client_instance))
        client_instance.get_object.assert_called_once_with('container',
                                                           'object')

    def test_validate_job_binary_url(self):

        job_binary = mock.Mock()

        # bad swift url should raise an exception
        job_binary.url = 'notswift://container/object'
        with testtools.ExpectedException(ex.BadJobBinaryException):
            self.i_s._validate_job_binary_url(job_binary)

        # specifying only a container should raise an exception
        job_binary.url = 'swift://container'
        with testtools.ExpectedException(ex.BadJobBinaryException):
            self.i_s._validate_job_binary_url(job_binary)

        job_binary.url = 'swift://container/path'
        self.i_s._validate_job_binary_url(job_binary)

    @mock.patch('sahara.service.edp.job_binaries.swift.implementation.'
                'SwiftType._get_raw_data')
    @mock.patch('sahara.utils.openstack.swift.client')
    def test_get_raw_data(self, swift_client, get_raw_data):
        client_instance = mock.Mock()
        swift_client.return_value = client_instance

        job_binary = mock.Mock()
        job_binary.url = 'swift://container/object'

        # embedded credentials
        job_binary.extra = dict(user='test', password='secret')
        self.i_s.get_raw_data(job_binary)
        swift_client.assert_called_with(username='test',
                                        password='secret')
        get_raw_data.assert_called_with(job_binary, client_instance)

        # proxy configs should override embedded credentials
        proxy_configs = dict(proxy_username='proxytest',
                             proxy_password='proxysecret',
                             proxy_trust_id='proxytrust')
        self.i_s.get_raw_data(job_binary, proxy_configs=proxy_configs)
        swift_client.assert_called_with(username='proxytest',
                                        password='proxysecret',
                                        trust_id='proxytrust')
        get_raw_data.assert_called_with(job_binary, client_instance)

    @mock.patch('sahara.utils.openstack.base.url_for')
    @mock.patch('sahara.context.ctx')
    @mock.patch(
        'sahara.service.edp.job_binaries.swift.'
        'implementation.SwiftType._get_raw_data')
    @mock.patch('swiftclient.Connection')
    def test_get_raw_data_with_context(self, swift_client, _get_raw_data, ctx,
                                       url_for):
        client_instance = mock.Mock()
        swift_client.return_value = client_instance
        test_context = mock.Mock()
        test_context.auth_token = 'testtoken'
        test_context.auth_plugin = None
        ctx.return_value = test_context
        url_for.return_value = 'url_for'
        job_binary = mock.Mock()
        job_binary.url = 'swift://container/object'

        job_binary.extra = dict(user='test', password='secret')
        self.i_s.get_raw_data(job_binary, with_context=True)
        self.assertEqual([mock.call(
            auth_version='3',
            cacert=None, insecure=False,
            max_backoff=10,
            preauthtoken='testtoken',
            preauthurl='url_for', retries=5,
            retry_on_ratelimit=True, starting_backoff=10)],
            swift_client.call_args_list)
        _get_raw_data.assert_called_with(job_binary, client_instance)



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
v2 API interface package

This package contains the endpoint definitions for the version 2 API.
The modules in this package are named in accordance with the top-level
resource primitives they represent.

This module provides a convenience function to register all the
endpoint blueprints to the ``/v2`` root.

When creating new endpoint modules, the following steps should be taken
to ensure they are properly registered with the Flask application:
  * create the module file with a name that indicates its endpoint
  * add a sahara.utils.api.RestV2 blueprint object
  * add an import to this module (__init__.py)
  * add a registration line for the new endpoint to the
    register_blueprint function
"""

from sahara.api.v2 import cluster_templates
from sahara.api.v2 import clusters
from sahara.api.v2 import data_sources
from sahara.api.v2 import images
from sahara.api.v2 import job_binaries
from sahara.api.v2 import job_templates
from sahara.api.v2 import job_types
from sahara.api.v2 import jobs
from sahara.api.v2 import node_group_templates
from sahara.api.v2 import plugins


def register_blueprints(app, url_prefix):
    """Register the v2 endpoints with a Flask application

    This function will take a Flask application object and register all
    the v2 endpoints. Register blueprints here when adding new endpoint
    modules.

    :param app: A Flask application object to register blueprints on
    :param url_prefix: The url prefix for the blueprints
    """
    app.register_blueprint(cluster_templates.rest, url_prefix=url_prefix)
    app.register_blueprint(clusters.rest, url_prefix=url_prefix)
    app.register_blueprint(data_sources.rest, url_prefix=url_prefix)
    app.register_blueprint(images.rest, url_prefix=url_prefix)
    app.register_blueprint(job_binaries.rest, url_prefix=url_prefix)
    app.register_blueprint(jobs.rest, url_prefix=url_prefix)
    app.register_blueprint(job_types.rest, url_prefix=url_prefix)
    app.register_blueprint(job_templates.rest, url_prefix=url_prefix)
    app.register_blueprint(node_group_templates.rest, url_prefix=url_prefix)
    app.register_blueprint(plugins.rest, url_prefix=url_prefix)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import manilaclient.client as manila_client
try:
    from manilaclient.common.apiclient import exceptions as manila_ex
except ImportError:
    from manilaclient.openstack.common.apiclient import exceptions as manila_ex
from oslo_config import cfg

from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.utils.openstack import base


opts = [
    cfg.StrOpt('api_version', default='1',
               help='Version of the manila API to use.'),
    cfg.BoolOpt('api_insecure',
                default=True,
                help='Allow to perform insecure SSL requests to manila.'),
    cfg.StrOpt('ca_file',
               help='Location of ca certificates file to use for manila '
                    'client requests.')
]

manila_group = cfg.OptGroup(name='manila',
                            title='Manila client options')

CONF = cfg.CONF
CONF.register_group(manila_group)
CONF.register_opts(opts, group=manila_group)

MANILA_PREFIX = "manila://"


def client():
    ctx = context.ctx()
    args = {
        'username': ctx.username,
        'project_name': ctx.tenant_name,
        'project_id': ctx.tenant_id,
        'input_auth_token': context.get_auth_token(),
        'auth_url': base.retrieve_auth_url(),
        'service_catalog_url': base.url_for(ctx.service_catalog, 'share'),
        'ca_cert': CONF.manila.ca_file,
        'insecure': CONF.manila.api_insecure
    }
    return manila_client.Client(CONF.manila.api_version, **args)


def get_share(client_instance, share_id, raise_on_error=False):
    try:
        return client_instance.shares.get(share_id)
    except manila_ex.NotFound:
        if raise_on_error:
            raise ex.NotFoundException(
                share_id, _("Share with id %s was not found."))
        else:
            return None



# Copyright (c) 2013 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import mock
from oslo_utils import timeutils
from oslo_utils import uuidutils
import testtools

from sahara import exceptions as ex
from sahara.service.api import v10 as api
from sahara.service.validations.edp import job_execution as je
from sahara.service.validations.edp import job_execution_schema as je_schema
from sahara.tests.unit.service.validation import utils as u
from sahara.tests.unit import testutils as tu
from sahara.utils import edp


def wrap_it(data):
    je.check_job_execution(data, 0)


class TestJobExecCreateValidation(u.ValidationTestCase):

    def setUp(self):
        super(TestJobExecCreateValidation, self).setUp()
        self._create_object_fun = wrap_it
        self.scheme = je_schema.JOB_EXEC_SCHEMA
        api.plugin_base.setup_plugins()

    @mock.patch('sahara.conductor.api.LocalApi.cluster_get')
    @mock.patch('sahara.conductor.api.LocalApi.data_source_get')
    @mock.patch('sahara.conductor.api.LocalApi.job_get')
    def test_streaming(self, get_job, get_data_source, get_cluster):
        get_job.return_value = mock.Mock(
            type=edp.JOB_TYPE_MAPREDUCE_STREAMING, libs=[], interface=[])

        ds1_id = uuidutils.generate_uuid()
        ds2_id = uuidutils.generate_uuid()

        data_sources = {
            ds1_id: mock.Mock(type="swift", url="http://swift/test"),
            ds2_id: mock.Mock(type="swift", url="http://swift/test2"),
        }

        get_data_source.side_effect = lambda ctx, x: data_sources[x]

        ng = tu.make_ng_dict('master', 42, ['oozie'], 1,
                             instances=[tu.make_inst_dict('id', 'name')])
        get_cluster.return_value = tu.create_cluster("cluster", "tenant1",
                                                     "fake", "0.1", [ng])

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "input_id": ds1_id,
                "output_id": ds2_id,
                "job_configs": {"configs": {},
                                "params": {},
                                "args": [],
                                "job_execution_info": {}}
            },
            bad_req_i=(1, "INVALID_DATA",
                          "MapReduce.Streaming job "
                          "must specify streaming mapper "
                          "and reducer"))

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "input_id": ds1_id,
                "output_id": ds2_id,
                "job_configs": {
                    "configs": {
                        "edp.streaming.mapper": "/bin/cat",
                        "edp.streaming.reducer": "/usr/bin/wc"},
                    "params": {},
                    "job_execution_info": {},
                    "args": []}
            })

    @mock.patch('sahara.conductor.api.LocalApi.cluster_get')
    @mock.patch('sahara.conductor.api.LocalApi.data_source_get')
    @mock.patch('sahara.conductor.api.LocalApi.job_get')
    def test_data_sources_differ(self, get_job, get_data_source, get_cluster):
        get_job.return_value = mock.Mock(
            type=edp.JOB_TYPE_MAPREDUCE_STREAMING, libs=[], interface=[])

        ds1_id = uuidutils.generate_uuid()
        ds2_id = uuidutils.generate_uuid()

        data_sources = {
            ds1_id: mock.Mock(type="swift", url="http://swift/test"),
            ds2_id: mock.Mock(type="swift", url="http://swift/test2"),
        }

        get_data_source.side_effect = lambda ctx, x: data_sources[x]

        ng = tu.make_ng_dict('master', 42, ['oozie'], 1,
                             instances=[tu.make_inst_dict('id', 'name')])
        get_cluster.return_value = tu.create_cluster("cluster", "tenant1",
                                                     "fake", "0.1", [ng])

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "input_id": ds1_id,
                "output_id": ds2_id,
                "job_configs": {
                    "configs": {
                        "edp.streaming.mapper": "/bin/cat",
                        "edp.streaming.reducer": "/usr/bin/wc"},
                    "params": {},
                    "job_execution_info": {},
                    "args": []}
            })

        data_sources[ds2_id].url = "http://swift/test"

        err_msg = ("Provided input and output DataSources reference the "
                   "same location: %s" % data_sources[ds2_id].url)

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "input_id": ds1_id,
                "output_id": ds2_id,
                "job_configs": {
                    "configs": {
                        "edp.streaming.mapper": "/bin/cat",
                        "edp.streaming.reducer": "/usr/bin/wc"},
                    "params": {},
                    "job_execution_info": {},
                    "args": []}
            },
            bad_req_i=(1, "INVALID_DATA", err_msg))

    @mock.patch('sahara.conductor.api.LocalApi.cluster_get')
    @mock.patch('sahara.conductor.api.LocalApi.job_get')
    def test_check_edp_job_support_spark(self, get_job, get_cluster):
        # utils.start_patch will construct a vanilla cluster as a
        # default for get_cluster, but we want a Spark cluster.
        # So, we'll make our own.

        # Note that this means we cannot use assert_create_object_validation()
        # because it calls start_patch() and will override our setting
        job = mock.Mock(type=edp.JOB_TYPE_SPARK, mains=["main"], interface=[])
        get_job.return_value = job
        ng = tu.make_ng_dict('master', 42, [], 1,
                             instances=[tu.make_inst_dict('id', 'name')])
        get_cluster.return_value = tu.create_cluster("cluster", "tenant1",
                                                     "fake", "0.1", [ng])

        # Everything is okay, spark cluster supports EDP by default
        # because cluster requires a master and slaves >= 1
        wrap_it(data={"cluster_id": uuidutils.generate_uuid(),
                      "job_configs": {
                          "configs": {
                              "edp.java.main_class": "org.me.class"}}})

    @mock.patch('sahara.conductor.api.LocalApi.cluster_get')
    @mock.patch('sahara.conductor.api.LocalApi.job_get')
    def test_edp_main_class_java(self, job_get, cluster_get):
        job_get.return_value = mock.Mock(type=edp.JOB_TYPE_JAVA,
                                         interface=[])
        ng = tu.make_ng_dict('master', 42, ['namenode', 'oozie'], 1,
                             instances=[tu.make_inst_dict('id', 'name')])
        cluster_get.return_value = tu.create_cluster("cluster", "tenant1",
                                                     "fake", "0.1", [ng])

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "job_configs": {"configs": {},
                                "params": {},
                                "args": [],
                                "job_execution_info": {}}
            },
            bad_req_i=(1, "INVALID_DATA",
                          "%s job must "
                          "specify edp.java.main_class" % edp.JOB_TYPE_JAVA))

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "job_configs": {
                    "configs": {
                        "edp.java.main_class": ""},
                    "params": {},
                    "args": [],
                    "job_execution_info": {}}
            },
            bad_req_i=(1, "INVALID_DATA",
                          "%s job must "
                          "specify edp.java.main_class" % edp.JOB_TYPE_JAVA))

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "job_configs": {
                    "configs": {
                        "edp.java.main_class": "org.me.myclass"},
                    "params": {},
                    "job_execution_info": {},
                    "args": []}
            })

    @mock.patch('sahara.conductor.api.LocalApi.cluster_get')
    @mock.patch('sahara.conductor.api.LocalApi.job_get')
    def test_edp_main_class_spark(self, job_get, cluster_get):
        job_get.return_value = mock.Mock(type=edp.JOB_TYPE_SPARK,
                                         interface=[])
        ng = tu.make_ng_dict('master', 42, ['namenode'], 1,
                             instances=[tu.make_inst_dict('id', 'name')])
        cluster_get.return_value = tu.create_cluster("cluster", "tenant1",
                                                     "fake", "0.1", [ng])

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "job_configs": {"configs": {},
                                "params": {},
                                "args": [],
                                "job_execution_info": {}}
            },
            bad_req_i=(1, "INVALID_DATA",
                          "%s job must "
                          "specify edp.java.main_class" % edp.JOB_TYPE_SPARK))

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "job_configs": {
                    "configs": {
                        "edp.java.main_class": ""},
                    "params": {},
                    "args": [],
                    "job_execution_info": {}}
            },
            bad_req_i=(1, "INVALID_DATA",
                          "%s job must "
                          "specify edp.java.main_class" % edp.JOB_TYPE_SPARK))

        self._assert_create_object_validation(
            data={
                "cluster_id": uuidutils.generate_uuid(),
                "job_configs": {
                    "configs": {
                        "edp.java.main_class": "org.me.myclass"},
                    "params": {},
                    "job_execution_info": {},
                    "args": []}
            })

    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_invalid_start_time_in_job_execution_info(self, now_get):
        configs = {"start": "2015-07-21 14:32:52"}
        now = time.strptime("2015-07-22 14:39:14", "%Y-%m-%d %H:%M:%S")
        now = timeutils.datetime.datetime.fromtimestamp(time.mktime(now))
        now_get.return_value = now

        with testtools.ExpectedException(ex.InvalidJobExecutionInfoException):
            je.check_scheduled_job_execution_info(configs)


class TestJobExecUpdateValidation(u.ValidationTestCase):

    def setUp(self):
        super(TestJobExecUpdateValidation, self).setUp()
        self._create_object_fun = mock.Mock()
        self.scheme = je_schema.JOB_EXEC_UPDATE_SCHEMA

    def test_job_execution_update_types(self):
        data = {
            'is_public': False,
            'is_protected': False,
            'info': {
                'status': 'suspend'
            }
        }
        self._assert_types(data)

    def test_job_execution_update_nothing_required(self):
        self._assert_create_object_validation(
            data={
                'is_public': False,
                'is_protected': False,
                'info': {
                    'status': 'suspend'
                }
            }
        )

    @mock.patch('sahara.conductor.api.LocalApi.job_execution_get')
    def test_je_update_when_protected(self, get_je_p):

        job_exec = mock.Mock(id='123', tenant_id='tenant_1', is_protected=True)
        get_je_p.return_value = job_exec

        # job execution can't be updated if it's marked as protected
        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                je.check_job_execution_update(job_exec, {'job_configs': {}})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e
        # job execution can be updated because is_protected flag was
        # set to False
        je.check_job_execution_update(
            job_exec, {'is_protected': False, 'job_configs': {}})

    @mock.patch('sahara.conductor.api.LocalApi.job_execution_get')
    def test_public_je_cancel_delete_from_another_tenant(self, get_je_p):

        job_exec = mock.Mock(id='123', tenant_id='tenant2', is_protected=False,
                             is_public=True)
        get_je_p.return_value = job_exec

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                je.check_job_execution_update(
                    job_exec, data={'is_public': False})
            except ex.UpdateFailedException as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e


class TestJobExecutionCancelDeleteValidation(u.ValidationTestCase):
    def setUp(self):
        super(TestJobExecutionCancelDeleteValidation, self).setUp()
        self.setup_context(tenant_id='tenant1')

    @mock.patch('sahara.conductor.api.LocalApi.job_execution_get')
    def test_je_cancel_delete_when_protected(self, get_je_p):

        job_exec = mock.Mock(id='123', tenant_id='tenant1', is_protected=True)
        get_je_p.return_value = job_exec

        with testtools.ExpectedException(ex.CancelingFailed):
            try:
                je.check_job_execution_cancel(job_exec)
            except ex.CancelingFailed as e:
                self.assert_protected_resource_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                je.check_job_execution_delete(job_exec)
            except ex.DeletionFailed as e:
                self.assert_protected_resource_exception(e)
                raise e

    @mock.patch('sahara.conductor.api.LocalApi.job_execution_get')
    def test_public_je_cancel_delete_from_another_tenant(self, get_je_p):

        job_exec = mock.Mock(id='123', tenant_id='tenant2', is_protected=False,
                             is_public=True)
        get_je_p.return_value = job_exec

        with testtools.ExpectedException(ex.CancelingFailed):
            try:
                je.check_job_execution_cancel(job_exec)
            except ex.CancelingFailed as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                je.check_job_execution_delete(job_exec)
            except ex.DeletionFailed as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e



# Copyright (c) 2017 Massachusetts Open Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from sahara.service.edp import s3_common


def get_raw_data(job_binary):
    return s3_common.get_raw_job_binary_data(job_binary)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jsonschema
from oslo_utils import uuidutils
import testtools

from sahara.utils import api_validator


def _validate(schema, data):
    validator = api_validator.ApiValidator(schema)
    validator.validate(data)


class ApiValidatorTest(testtools.TestCase):
    def _validate_success(self, schema, data):
        return _validate(schema, data)

    def _validate_failure(self, schema, data):
        self.assertRaises(jsonschema.ValidationError, _validate, schema, data)

    def test_validate_required(self):
        schema = {
            "type": "object",
            "properties": {
                "prop-1": {
                    "type": "string",
                },
            },
        }

        self._validate_success(schema, {
            "prop-1": "asd",
        })
        self._validate_success(schema, {
            "prop-2": "asd",
        })

        schema["required"] = ["prop-1"]

        self._validate_success(schema, {
            "prop-1": "asd",
        })
        self._validate_failure(schema, {
            "prop-2": "asd",
        })

    def test_validate_additionalProperties(self):
        schema = {
            "type": "object",
            "properties": {
                "prop-1": {
                    "type": "string",
                },
            },
            "required": ["prop-1"]
        }

        self._validate_success(schema, {
            "prop-1": "asd",
        })
        self._validate_success(schema, {
            "prop-1": "asd",
            "prop-2": "asd",
        })

        schema["additionalProperties"] = True

        self._validate_success(schema, {
            "prop-1": "asd",
        })
        self._validate_success(schema, {
            "prop-1": "asd",
            "prop-2": "asd",
        })

        schema["additionalProperties"] = False

        self._validate_success(schema, {
            "prop-1": "asd",
        })
        self._validate_failure(schema, {
            "prop-1": "asd",
            "prop-2": "asd",
        })

    def test_validate_string(self):
        schema = {
            "type": "string",
        }

        self._validate_success(schema, "asd")
        self._validate_success(schema, "")
        self._validate_failure(schema, 1)
        self._validate_failure(schema, 1.5)
        self._validate_failure(schema, True)

    def test_validate_string_with_length(self):
        schema = {
            "type": "string",
            "minLength": 1,
            "maxLength": 10,
        }

        self._validate_success(schema, "a")
        self._validate_success(schema, "a" * 10)
        self._validate_failure(schema, "")
        self._validate_failure(schema, "a" * 11)

    def test_validate_integer(self):
        schema = {
            'type': 'integer',
        }

        self._validate_success(schema, 0)
        self._validate_success(schema, 1)
        self._validate_failure(schema, "1")
        self._validate_failure(schema, "a")
        self._validate_failure(schema, True)

    def test_validate_integer_w_range(self):
        schema = {
            'type': 'integer',
            'minimum': 1,
            'maximum': 10,
        }

        self._validate_success(schema, 1)
        self._validate_success(schema, 10)
        self._validate_failure(schema, 0)
        self._validate_failure(schema, 11)

    def test_validate_uuid(self):
        schema = {
            "type": "string",
            "format": "uuid",
        }

        id = uuidutils.generate_uuid()

        self._validate_success(schema, id)
        self._validate_success(schema, id.replace("-", ""))

    def test_validate_valid_name(self):
        schema = {
            "type": "string",
            "format": "valid_name",
        }

        self._validate_success(schema, "abcd")
        self._validate_success(schema, "abcd123")
        self._validate_success(schema, "abcd-123")
        self._validate_success(schema, "abcd_123")
        self._validate_failure(schema, "_123")
        self._validate_success(schema, "a" * 64)
        self._validate_failure(schema, "")
        self._validate_success(schema, "hadoop-examples-2.6.0.jar")
        self._validate_success(schema, "hadoop-examples-2.6.0")
        self._validate_success(schema, "hadoop-examples-2.6.0.")
        self._validate_success(schema, "1")
        self._validate_success(schema, "1a")
        self._validate_success(schema, "a1")
        self._validate_success(schema, "A1")
        self._validate_success(schema, "A1B")
        self._validate_success(schema, "a.b")
        self._validate_success(schema, "a..b")
        self._validate_success(schema, "a._.b")
        self._validate_success(schema, "a_")
        self._validate_success(schema, "a-b-001")
        self._validate_failure(schema, "-aaaa-bbbb")
        self._validate_failure(schema, ".aaaa-bbbb")

        self._validate_failure(schema, None)
        self._validate_failure(schema, 1)
        self._validate_failure(schema, ["1"])

    def test_validate_valid_keypair_name(self):
        schema = {
            "type": "string",
            "format": "valid_keypair_name",
        }

        self._validate_success(schema, "abcd")
        self._validate_success(schema, "abcd123")
        self._validate_success(schema, "abcd-123")
        self._validate_success(schema, "abcd_123")
        self._validate_success(schema, "_123")
        self._validate_success(schema, "a" * 64)
        self._validate_failure(schema, "")
        self._validate_failure(schema, "hadoop-examples-2.6.0.jar")
        self._validate_failure(schema, "hadoop-examples-2.6.0")
        self._validate_failure(schema, "hadoop-examples-2.6.0.")
        self._validate_success(schema, "1")
        self._validate_success(schema, "1a")
        self._validate_success(schema, "a1")
        self._validate_success(schema, "A1")
        self._validate_success(schema, "A1B")
        self._validate_failure(schema, "a.b")
        self._validate_failure(schema, "a..b")
        self._validate_failure(schema, "a._.b")
        self._validate_success(schema, "a_")
        self._validate_success(schema, "a-b-001")
        self._validate_success(schema, "-aaaa-bbbb")
        self._validate_success(schema, "-aaaa bbbb")
        self._validate_success(schema, " -aaaa bbbb")
        self._validate_failure(schema, ".aaaa-bbbb")

        self._validate_failure(schema, None)
        self._validate_failure(schema, 1)
        self._validate_failure(schema, ["1"])

    def test_validate_valid_name_hostname(self):
        schema = {
            "type": "string",
            "format": "valid_name_hostname",
            "minLength": 1,
        }

        self._validate_success(schema, "abcd")
        self._validate_success(schema, "abcd123")
        self._validate_success(schema, "abcd-123")
        self._validate_failure(schema, "abcd_123")
        self._validate_failure(schema, "_123")
        self._validate_success(schema, "a" * 64)
        self._validate_failure(schema, "")
        self._validate_failure(schema, "hadoop-examples-2.6.0.jar")
        self._validate_failure(schema, "hadoop-examples-2.6.0")
        self._validate_failure(schema, "hadoop-examples-2.6.0.")
        self._validate_failure(schema, "1")
        self._validate_failure(schema, "1a")
        self._validate_success(schema, "a1")
        self._validate_success(schema, "A1")
        self._validate_success(schema, "A1B")
        self._validate_success(schema, "aB")
        self._validate_success(schema, "a.b")
        self._validate_failure(schema, "a..b")
        self._validate_failure(schema, "a._.b")
        self._validate_failure(schema, "a_")
        self._validate_success(schema, "a-b-001")

        self._validate_failure(schema, None)
        self._validate_failure(schema, 1)
        self._validate_failure(schema, ["1"])

    def test_validate_hostname(self):
        schema = {
            "type": "string",
            "format": "hostname",
        }

        self._validate_success(schema, "abcd")
        self._validate_success(schema, "abcd123")
        self._validate_success(schema, "abcd-123")
        self._validate_failure(schema, "abcd_123")
        self._validate_failure(schema, "_123")
        self._validate_failure(schema, "a" * 64)
        self._validate_failure(schema, "")

    def test_validate_configs(self):
        schema = {
            "type": "object",
            "properties": {
                "configs": {
                    "type": "configs",
                }
            },
            "additionalProperties": False
        }

        self._validate_success(schema, {
            "configs": {
                "at-1": {
                    "c-1": "c",
                    "c-2": 1,
                    "c-3": True,
                },
                "at-2": {
                    "c-4": "c",
                    "c-5": 1,
                    "c-6": True,
                },
            },
        })

        self._validate_failure(schema, {
            "configs": {
                "at-1": {
                    "c-1": 1.5
                },
            }
        })

        self._validate_failure(schema, {
            "configs": {
                1: {
                    "c-1": "c"
                },
            }
        })

        self._validate_failure(schema, {
            "configs": {
                "at-1": {
                    1: "asd",
                },
            }
        })

        self._validate_failure(schema, {
            "configs": {
                "at-1": [
                    "a", "b", "c",
                ],
            }
        })

    def test_validate_flavor(self):
        schema = {
            'type': "flavor",
        }

        self._validate_success(schema, 0)
        self._validate_success(schema, 1)
        self._validate_success(schema, "0")
        self._validate_success(schema, "1")
        self._validate_success(schema, uuidutils.generate_uuid())
        self._validate_failure(schema, True)
        self._validate_failure(schema, 0.1)
        self._validate_failure(schema, "0.1")
        self._validate_failure(schema, "asd")



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from sahara import context
from sahara import exceptions as ex
from sahara.utils import cluster as cluster_utils

LOG = logging.getLogger(__name__)

# set 3 hours timeout by default
DEFAULT_TIMEOUT = 10800
DEFAULT_SLEEP_TIME = 5

timeouts_opts = [
    # engine opts
    cfg.IntOpt('ips_assign_timeout',
               default=DEFAULT_TIMEOUT,
               help="Assign IPs timeout, in seconds"),
    cfg.IntOpt('wait_until_accessible',
               default=DEFAULT_TIMEOUT,
               help="Wait for instance accessibility, in seconds"),

    # direct engine opts
    cfg.IntOpt('delete_instances_timeout',
               default=DEFAULT_TIMEOUT,
               help="Wait for instances to be deleted, in seconds"),

    # volumes opts
    cfg.IntOpt(
        'detach_volume_timeout', default=300,
        help='Timeout for detaching volumes from instance, in seconds'),
]

timeouts = cfg.OptGroup(name='timeouts',
                        title='Sahara timeouts')

CONF = cfg.CONF
CONF.register_group(timeouts)
CONF.register_opts(timeouts_opts, group=timeouts)


def _get_consumed(started_at):
    return timeutils.delta_seconds(started_at, timeutils.utcnow())


def _get_current_value(cluster, option):
    option_target = option.applicable_target
    conf = cluster.cluster_configs
    if option_target in conf and option.name in conf[option_target]:
        return conf[option_target][option.name]
    return option.default_value


def poll(get_status, kwargs=None, args=None, operation_name=None,
         timeout_name=None, timeout=DEFAULT_TIMEOUT, sleep=DEFAULT_SLEEP_TIME,
         exception_strategy='raise'):
    """This util poll status of object obj during some timeout.

    :param get_status: function, which return current status of polling
    as Boolean
    :param kwargs: keyword arguments of function get_status
    :param operation_name: name of polling process
    :param timeout_name: name of timeout option
    :param timeout: value of timeout in seconds. By default, it equals to
    3 hours
    :param sleep: duration between two consecutive executions of
    get_status function
    :param exception_strategy: possible values ('raise', 'mark_as_true',
    'mark_as_false'). If exception_strategy is 'raise' exception would be
    raised. If exception_strategy is 'mark_as_true', return value of
    get_status would marked as True, and in case of 'mark_as_false' - False.
    By default it's 'raise'.
    """
    start_time = timeutils.utcnow()
    # We shouldn't raise TimeoutException if incorrect timeout specified and
    # status is ok now. In such way we should execute get_status at least once.
    at_least_once = True
    if not kwargs:
        kwargs = {}
    if not args:
        args = ()

    while at_least_once or _get_consumed(start_time) < timeout:
        at_least_once = False
        try:
            status = get_status(*args, **kwargs)
        except BaseException:
            if exception_strategy == 'raise':
                raise
            elif exception_strategy == 'mark_as_true':
                status = True
            else:
                status = False

        if status:
            operation = "Operation"
            if operation_name:
                operation = "Operation with name {op_name}".format(
                    op_name=operation_name)
            LOG.debug(
                '{operation_desc} was executed successfully in timeout '
                '{timeout}'
                .format(operation_desc=operation, timeout=timeout))
            return

        context.sleep(sleep)
    raise ex.TimeoutException(timeout, operation_name, timeout_name)


def plugin_option_poll(cluster, get_status, option, operation_name, sleep_time,
                       kwargs):

    def _get(n_cluster, n_kwargs):
        if not cluster_utils.check_cluster_exists(n_cluster):
            return True
        return get_status(**n_kwargs)

    poll_description = {
        'get_status': _get,
        'kwargs': {'n_cluster': cluster, 'n_kwargs': kwargs},
        'timeout': _get_current_value(cluster, option),
        'operation_name': operation_name,
        'sleep': sleep_time,
        'timeout_name': option.name
    }

    poll(**poll_description)


def poll_status(option, operation_name, sleep):
    def decorator(f):
        @functools.wraps(f)
        def handler(*args, **kwargs):
            poll_description = {
                'get_status': f,
                'kwargs': kwargs,
                'args': args,
                'timeout': getattr(CONF.timeouts, option),
                'operation_name': operation_name,
                'timeout_name': option,
                'sleep': sleep,
            }
            poll(**poll_description)
        return handler
    return decorator



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara import context as ctx
from sahara.plugins import base as pb
from sahara.service.edp.job_utils import ds_manager
from sahara.service.edp.oozie import engine as oe
from sahara.service.edp.oozie.engine import jb_manager
from sahara.tests.unit import base
from sahara.tests.unit.service.edp import edp_test_utils as u
from sahara.utils import edp


class TestOozieEngine(base.SaharaTestCase):
    def setUp(self):
        super(TestOozieEngine, self).setUp()
        self.override_config('plugins', ['fake'])
        pb.setup_plugins()
        jb_manager.setup_job_binaries()
        ds_manager.setup_data_sources()

    def test_get_job_status(self):
        oje = FakeOozieJobEngine(u.create_cluster())
        client_class = mock.MagicMock()
        client_class.add_job = mock.MagicMock(return_value=1)
        client_class.get_job_info = mock.MagicMock(
            return_value={'status': 'PENDING'})
        oje.get_client = mock.MagicMock(return_value=client_class)

        _, job_exec = u.create_job_exec(edp.JOB_TYPE_PIG)
        self.assertIsNone(oje.get_job_status(job_exec))

        job_exec.engine_job_id = 1
        self.assertEqual({'status': 'PENDING'}, oje.get_job_status(job_exec))

    def test_add_postfix(self):
        oje = FakeOozieJobEngine(u.create_cluster())

        self.override_config("job_workflow_postfix", 'caba')
        res = oje._add_postfix('aba')
        self.assertEqual("aba/caba/", res)

        self.override_config("job_workflow_postfix", '')
        res = oje._add_postfix('aba')
        self.assertEqual("aba/", res)

    def test_get_oozie_job_params(self):
        oje = FakeOozieJobEngine(u.create_cluster())
        oozie_params = {'oozie.libpath': '/mylibpath',
                        'oozie.wf.application.path': '/wrong'}
        scheduled_params = {'start': '2015-06-10T06:05Z',
                            'end': '2015-06-10T06:50Z',
                            'frequency': '10'}
        job_dir = '/job_dir'
        job_execution_type = 'workflow'
        job_params = oje._get_oozie_job_params('hadoop',
                                               '/tmp', oozie_params, True,
                                               scheduled_params, job_dir,
                                               job_execution_type)
        self.assertEqual('http://localhost:50030', job_params["jobTracker"])
        self.assertEqual('hdfs://localhost:8020', job_params["nameNode"])
        self.assertEqual('hadoop', job_params["user.name"])
        self.assertEqual('hdfs://localhost:8020/tmp',
                         job_params['oozie.wf.application.path'])
        self.assertEqual("/mylibpath,hdfs://localhost:8020/user/"
                         "sahara-hbase-lib", job_params['oozie.libpath'])

        # Make sure this doesn't raise an exception
        job_params = oje._get_oozie_job_params('hadoop',
                                               '/tmp', {}, True)
        self.assertEqual("hdfs://localhost:8020/user/"
                         "sahara-hbase-lib", job_params['oozie.libpath'])

        job_execution_type = 'scheduled'
        job_params = oje._get_oozie_job_params('hadoop',
                                               '/tmp', oozie_params, True,
                                               scheduled_params, job_dir,
                                               job_execution_type)
        for i in ["start", "end", "frequency"]:
            self.assertEqual(scheduled_params[i], job_params[i])

    @mock.patch('sahara.utils.remote.get_remote')
    @mock.patch('sahara.utils.ssh_remote.InstanceInteropHelper')
    @mock.patch('sahara.conductor.API.job_binary_internal_get_raw_data')
    def test_hdfs_upload_job_files(self, conductor_raw_data, remote_class,
                                   remote):
        remote_class.__exit__.return_value = 'closed'
        remote.return_value = remote_class
        conductor_raw_data.return_value = 'ok'

        oje = FakeOozieJobEngine(u.create_cluster())
        oje._prepare_job_binaries = mock.Mock()

        job, _ = u.create_job_exec(edp.JOB_TYPE_PIG)
        res = oje._upload_job_files_to_hdfs(mock.Mock(), 'job_prefix', job, {})
        self.assertEqual(['/tmp/script.pig'], res)

        job, _ = u.create_job_exec(edp.JOB_TYPE_MAPREDUCE)
        res = oje._upload_job_files_to_hdfs(mock.Mock(), 'job_prefix', job, {})
        self.assertEqual(['/tmp/main.jar'], res)

    @mock.patch('sahara.utils.remote.get_remote')
    def test_upload_workflow_file(self, remote_get):
        oje = FakeOozieJobEngine(u.create_cluster())
        remote_class = mock.MagicMock()
        remote_class.__exit__.return_value = 'closed'
        remote_get.return_value = remote_class
        res = oje._upload_workflow_file(remote_get, "test", "hadoop.xml",
                                        'hdfs')
        self.assertEqual("test/workflow.xml", res)

    @mock.patch('sahara.utils.remote.get_remote')
    def test_upload_coordinator_file(self, remote_get):
        oje = FakeOozieJobEngine(u.create_cluster())
        remote_class = mock.MagicMock()
        remote_class.__exit__.return_value = 'closed'
        remote_get.return_value = remote_class
        res = oje._upload_coordinator_file(remote_get, "test", "hadoop.xml",
                                           'hdfs')
        self.assertEqual("test/coordinator.xml", res)

    @mock.patch('sahara.utils.remote.get_remote')
    def test_hdfs_create_workflow_dir(self, remote):
        remote_class = mock.MagicMock()
        remote_class.__exit__.return_value = 'closed'
        remote.return_value = remote_class

        oje = FakeOozieJobEngine(u.create_cluster())
        job, _ = u.create_job_exec(edp.JOB_TYPE_PIG)
        res = oje._create_hdfs_workflow_dir(mock.Mock(), job)
        self.assertIn('/user/hadoop/special_name/', res)

    def test__resolve_external_hdfs_urls(self):

        oje = FakeOozieJobEngine(u.create_cluster())
        job_configs = {
            "configs": {
                "mapred.map.tasks": "1",
                "hdfs1": "hdfs://localhost/hdfs1"},
            "args": ["hdfs://localhost/hdfs3", "10"],
            "params": {
                "param1": "10",
                "param2": "hdfs://localhost/hdfs2"
            }
        }

        expected_external_hdfs_urls = ['hdfs://localhost/hdfs1',
                                       'hdfs://localhost/hdfs2',
                                       'hdfs://localhost/hdfs3']

        external_hdfs_urls = oje._resolve_external_hdfs_urls(job_configs)

        self.assertEqual(expected_external_hdfs_urls, external_hdfs_urls)

    @mock.patch('sahara.service.edp.oozie.oozie.OozieClient.get_job_info')
    @mock.patch('sahara.service.edp.oozie.oozie.OozieClient.kill_job')
    def test_cancel_job(self, kill_get, info_get):
        info_get.return_value = {}
        oje = FakeOozieJobEngine(u.create_cluster())
        _, job_exec = u.create_job_exec(edp.JOB_TYPE_PIG)

        # test cancel job without engine_job_id
        job_exec.engine_job_id = None
        oje.cancel_job(job_exec)
        self.assertEqual(0, kill_get.call_count)

        # test cancel job with engine_job_id
        job_exec.engine_job_id = 123
        oje.cancel_job(job_exec)
        self.assertEqual(1, kill_get.call_count)

    @mock.patch('sahara.service.edp.job_utils.prepare_cluster_for_ds')
    @mock.patch('sahara.service.edp.job_utils._get_data_source_urls')
    @mock.patch('sahara.service.edp.oozie.workflow_creator.'
                'workflow_factory.get_workflow_xml')
    @mock.patch('sahara.utils.remote.get_remote')
    @mock.patch('sahara.conductor.API.job_execution_update')
    @mock.patch('sahara.conductor.API.data_source_get')
    @mock.patch('sahara.conductor.API.job_get')
    def test_prepare_run_job(self, job, data_source, update,
                             remote, wf_factory, get_ds_urls,
                             prepare_cluster):
        wf_factory.return_value = mock.MagicMock()

        remote_class = mock.MagicMock()
        remote_class.__exit__.return_value = 'closed'
        remote.return_value = remote_class

        job_class = mock.MagicMock()
        job_class.name = "myJob"
        job.return_value = job_class

        source = mock.MagicMock()
        source.url = "localhost"

        get_ds_urls.return_value = ('url', 'url')

        data_source.return_value = source
        oje = FakeOozieJobEngine(u.create_cluster())
        _, job_exec = u.create_job_exec(edp.JOB_TYPE_PIG)
        update.return_value = job_exec

        res = oje._prepare_run_job(job_exec)
        self.assertEqual(ctx.ctx(), res['context'])
        self.assertEqual('hadoop', res['hdfs_user'])
        self.assertEqual(job_exec, res['job_execution'])
        self.assertEqual({}, res['oozie_params'])

    @mock.patch('sahara.service.edp.job_utils.prepare_cluster_for_ds')
    @mock.patch('sahara.service.edp.job_utils._get_data_source_urls')
    @mock.patch('sahara.service.edp.oozie.workflow_creator.'
                'workflow_factory.get_workflow_xml')
    @mock.patch('sahara.utils.remote.get_remote')
    @mock.patch('sahara.conductor.API.job_execution_update')
    @mock.patch('sahara.conductor.API.data_source_get')
    @mock.patch('sahara.conductor.API.job_get')
    @mock.patch('sahara.conductor.API.job_execution_get')
    def test_run_job(self, exec_get, job, data_source,
                     update, remote, wf_factory, get_ds_urls,
                     prepare_cluster):
        wf_factory.return_value = mock.MagicMock()
        remote_class = mock.MagicMock()
        remote_class.__exit__.return_value = 'closed'
        remote.return_value = remote_class

        job_class = mock.MagicMock()
        job.return_value = job_class
        job.name = "myJob"

        source = mock.MagicMock()
        source.url = "localhost"
        data_source.return_value = source

        get_ds_urls.return_value = ('url', 'url')

        oje = FakeOozieJobEngine(u.create_cluster())
        client_class = mock.MagicMock()
        client_class.add_job = mock.MagicMock(return_value=1)
        client_class.get_job_info = mock.MagicMock(
            return_value={'status': 'PENDING'})
        oje.get_client = mock.MagicMock(return_value=client_class)

        _, job_exec = u.create_job_exec(edp.JOB_TYPE_PIG)
        update.return_value = job_exec

        self.assertEqual((1, 'PENDING', None), oje.run_job(job_exec))


class FakeOozieJobEngine(oe.OozieJobEngine):
    def get_hdfs_user(self):
        return 'hadoop'

    def create_hdfs_dir(self, remote, dir_name):
        return

    def get_oozie_server_uri(self, cluster):
        return 'http://localhost:11000/oozie'

    def get_oozie_server(self, cluster):
        return None

    def get_name_node_uri(self, cluster):
        return 'hdfs://localhost:8020'

    def get_resource_manager_uri(self, cluster):
        return 'http://localhost:50030'



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara import context
from sahara.i18n import _
from sahara.plugins import exceptions as pex
from sahara.plugins.fake import edp_engine
from sahara.plugins import kerberos as krb
from sahara.plugins import provisioning as p
from sahara.plugins import utils as plugin_utils


class FakePluginProvider(p.ProvisioningPluginBase):

    def get_title(self):
        return "Fake Plugin"

    def get_description(self):
        return _("It's a fake plugin that aimed to work on the CirrOS images. "
                 "It doesn't install Hadoop. It's needed to be able to test "
                 "provisioning part of Sahara codebase itself.")

    def get_versions(self):
        return ["0.1"]

    def get_labels(self):
        return {
            'plugin_labels': {
                'enabled': {'status': True},
                'hidden': {'status': True},
            },
            'version_labels': {
                '0.1': {'enabled': {'status': True}}
            }
        }

    def get_node_processes(self, hadoop_version):
        return {
            "HDFS": ["namenode", "datanode"],
            "MapReduce": ["tasktracker", "jobtracker"],
            "Kerberos": [],
        }

    def get_configs(self, hadoop_version):
        # returning kerberos configs
        return krb.get_config_list()

    def configure_cluster(self, cluster):
        with context.ThreadGroup() as tg:
            for instance in plugin_utils.get_instances(cluster):
                tg.spawn('fake-write-%s' % instance.id,
                         self._write_ops, instance)

    def start_cluster(self, cluster):
        self.deploy_kerberos(cluster)
        with context.ThreadGroup() as tg:
            for instance in plugin_utils.get_instances(cluster):
                tg.spawn('fake-check-%s' % instance.id,
                         self._check_ops, instance)

    def deploy_kerberos(self, cluster):
        all_instances = plugin_utils.get_instances(cluster)
        namenodes = plugin_utils.get_instances(cluster, 'namenode')
        server = None
        if len(namenodes) > 0:
            server = namenodes[0]
        elif len(all_instances) > 0:
            server = all_instances[0]
        if server:
            krb.deploy_infrastructure(cluster, server)

    def scale_cluster(self, cluster, instances):
        with context.ThreadGroup() as tg:
            for instance in instances:
                tg.spawn('fake-scaling-%s' % instance.id,
                         self._all_check_ops, instance)

    def decommission_nodes(self, cluster, instances):
        pass

    def _write_ops(self, instance):
        with instance.remote() as r:
            # check typical SSH command
            r.execute_command('echo "Hello, world!"')
            # check write file
            data_1 = "sp@m"
            r.write_file_to('test_data', data_1, run_as_root=True)
            # check append file
            data_2 = " and eggs"
            r.append_to_file('test_data', data_2, run_as_root=True)
            # check replace string
            r.replace_remote_string('test_data', "eggs", "pony")

    def _check_ops(self, instance):
        expected_data = "sp@m and pony"
        with instance.remote() as r:
            actual_data = r.read_file_from('test_data', run_as_root=True)

            if actual_data.strip() != expected_data.strip():
                raise pex.HadoopProvisionError("ACTUAL:\n%s\nEXPECTED:\n%s" % (
                    actual_data, expected_data))

    def _all_check_ops(self, instance):
        self._write_ops(instance)
        self._check_ops(instance)

    def get_edp_engine(self, cluster, job_type):
        if job_type in edp_engine.FakeJobEngine.get_supported_job_types():
            return edp_engine.FakeJobEngine()

    def get_edp_job_types(self, versions=None):
        res = {}
        for vers in self.get_versions():
            if not versions or vers in versions:
                res[vers] = edp_engine.FakeJobEngine.get_supported_job_types()
        return res

    def get_edp_config_hints(self, job_type, version):
        if version in self.get_versions():
            return edp_engine.FakeJobEngine.get_possible_job_config(job_type)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from sahara import context
import sahara.tests.unit.conductor.base as test_base
from sahara.tests.unit.conductor.manager import test_edp


def _merge_dict(original, update):
    new = copy.deepcopy(original)
    new.update(update)
    return new


SAMPLE_JOB = _merge_dict(test_edp.SAMPLE_JOB, {
    "interface": [
        {
            "name": "Reducer Count",
            "mapping_type": "configs",
            "location": "mapred.reduce.tasks",
            "value_type": "number",
            "required": True,
            "default": "1"
        },
        {
            "name": "Input Path",
            "mapping_type": "params",
            "location": "INPUT",
            "value_type": "data_source",
            "required": False,
            "default": "hdfs://path"
        },
        {
            "name": "Positional Argument 2",
            "mapping_type": "args",
            "location": "1",
            "value_type": "string",
            "required": False,
            "default": "default"
        },
        {
            "name": "Positional Argument 1",
            "mapping_type": "args",
            "location": "0",
            "value_type": "string",
            "required": False,
            "default": "arg_1"
        },

    ]
})

SAMPLE_JOB_EXECUTION = _merge_dict(test_edp.SAMPLE_JOB, {
    "interface": {
        "Reducer Count": "2",
        "Positional Argument 2": "arg_2"
    },
    "job_configs": {"args": ["arg_3"], "configs": {"mapred.map.tasks": "3"}}
})


class JobExecutionTest(test_base.ConductorManagerTestCase):
    def test_interface_flows(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)
        arg_names = [arg['name'] for arg in job['interface']]
        self.assertEqual(arg_names, ["Reducer Count", "Input Path",
                                     "Positional Argument 2",
                                     "Positional Argument 1"])

        job_ex_input = copy.deepcopy(SAMPLE_JOB_EXECUTION)
        job_ex_input['job_id'] = job['id']

        self.api.job_execution_create(ctx, job_ex_input)

        lst = self.api.job_execution_get_all(ctx)
        self.assertEqual(1, len(lst))

        job_ex_result = lst[0]
        configs = {
            'configs': {'mapred.reduce.tasks': '2',
                        'mapred.map.tasks': '3'},
            'args': ['arg_1', 'arg_2', 'arg_3'],
            'params': {'INPUT': 'hdfs://path'}
        }
        self.assertEqual(configs, job_ex_result['job_configs'])
        self.api.job_execution_destroy(ctx, job_ex_result['id'])

        del job_ex_input['job_configs']
        self.api.job_execution_create(ctx, job_ex_input)

        lst = self.api.job_execution_get_all(ctx)
        self.assertEqual(1, len(lst))

        job_ex_result = lst[0]
        configs = {
            'configs': {'mapred.reduce.tasks': '2'},
            'args': ['arg_1', 'arg_2'],
            'params': {'INPUT': 'hdfs://path'}
        }
        self.assertEqual(configs, job_ex_result['job_configs'])
